@@ -4,19 +4,33 @@ import collections
 import sys
 import time
 import uuid
+from contextlib import nullcontext
 from datetime import datetime, timedelta
 from functools import cached_property
-from typing import TYPE_CHECKING, Dict, Iterable, List, Mapping, Optional, Tuple
+from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple
 
 from rich import box
 from rich.console import Group
 from rich.padding import Padding
 from rich.panel import Panel
-from rich.progress import BarColumn, MofNCompleteColumn, Progress, SpinnerColumn, TaskID, TextColumn, TimeElapsedColumn
+from rich.progress import (
+    BarColumn,
+    MofNCompleteColumn,
+    Progress,
+    SpinnerColumn,
+    Task,
+    TaskID,
+    TaskProgressColumn,
+    TextColumn,
+    TimeElapsedColumn,
+)
 from rich.style import Style
 from rich.table import Table
 from rich.text import Text
+from tenacity import Retrying, retry_if_exception_message, stop_after_attempt
+from tenacity.wait import wait_exponential_jitter
 
+import chalk
 from chalk import EnvironmentId
 from chalk._reporting.models import BatchOpKind, BatchOpStatus, BatchReport
 from chalk._reporting.rich.color import PASTELY_CYAN, SHADOWY_LAVENDER, UNDERLYING_CYAN
@@ -30,10 +44,6 @@ if TYPE_CHECKING:
 
 _logger = get_logger(__name__)
 
-_kind_to_display_type = {
-    BatchOpKind.RECOMPUTE: "Dataset Recomputation",
-    BatchOpKind.OFFLINE_QUERY: "Offline Query",
-}
 
 _MAX_ERRORS_TO_DISPLAY = 10
 _TTableRow = Tuple[str, str]
@@ -47,6 +57,27 @@ except ImportError:
     _running_in_jupyter_notebook = False
 
 
+class MofOptionalNCompleteColumn(MofNCompleteColumn):
+    """
+    Just like MofNCompleteColumn but renders just the completed count,
+    instead of `completed / ?` if there is no total
+    """
+
+    def render(self, task: Task) -> Text:
+        """Show completed/total."""
+        completed = int(task.completed)
+        if task.total:
+            total = int(task.total)
+            total_width = len(str(total))
+            fmt = f"{completed:{total_width}d}{self.separator}{total}"
+        else:
+            fmt = f"{completed}"
+        return Text(fmt, style="progress.download")
+
+
+OFFLINE_QUERY = "Offline Query"
+
+
 class ProgressService:
     def __init__(
         self,
@@ -72,6 +103,7 @@ class ProgressService:
         self._environment_id = environment_id
         self._operation_kind: Optional[BatchOpKind] = None
         self._num_computers = num_computers
+        self._shard_id = 0
         if show_progress and not (_running_in_jupyter_notebook or sys.stdout.isatty()):
             _logger.warning("Progress display is not supported in dumb terminals. Progress will not be shown.")
             show_progress = False
@@ -84,7 +116,7 @@ class ProgressService:
                     style=Style(color=SHADOWY_LAVENDER),
                     complete_style=Style(color=UNDERLYING_CYAN),
                 ),
-                MofNCompleteColumn(),
+                MofOptionalNCompleteColumn(),
                 TextColumn("runs", style=Style(color=PASTELY_CYAN)),
                 TextColumn("{task.description}"),
             )
@@ -101,16 +133,35 @@ class ProgressService:
                 if caller_method
                 else None
             )
+            if self._num_computers > 1:
+                self.shard_progress = Progress(
+                    TextColumn("{task.description}"),
+                    MofNCompleteColumn(),
+                    BarColumn(
+                        finished_style=Style(color=PASTELY_CYAN),
+                        style=Style(color=SHADOWY_LAVENDER),
+                        complete_style=Style(color=UNDERLYING_CYAN),
+                    ),
+                    TaskProgressColumn(),
+                )
+            else:
+                self.shard_progress = None
         else:
             self.resolver_progress = None
             self.main_progress = None
             self.explainer_text = None
+            self.shard_progress = None
+        self.show_progress = show_progress
 
     @cached_property
     def enclosure_panel(self) -> Panel | None:
         if self.main_progress is None or self.resolver_progress is None:
             return None
-        progress_bar_group = Group(self.main_progress, self.resolver_progress)
+        if self.shard_progress is None:
+            progress_bar_group = Group(self.main_progress, self.resolver_progress)
+        else:
+            blank_line = Padding(Text(""), pad=(0, 0, 0, 0))
+            progress_bar_group = Group(self.shard_progress, blank_line, self.main_progress, self.resolver_progress)
         panel_content_table = Table.grid()
         if self.explainer_text is not None:
             panel_content_table.add_row(Padding(self.explainer_text, (0, 0, 1, 0)))
@@ -131,7 +182,7 @@ class ProgressService:
                 short_name = fqn.split(".")[-1]
                 fqn_to_task_id[fqn] = self.resolver_progress.add_task(
                     description=f"[ {short_name} ]",
-                    total=resolver_report.progress.total,
+                    total=resolver_report.progress.total or None,
                 )
 
         for resolver_report in batch_report.resolvers:
@@ -143,14 +194,27 @@ class ProgressService:
                 resolver_task_id,
                 completed=rows_done_new,
                 # total might change as the same resolver gets invoked in a separate node
-                total=resolver_report.progress.total,
+                total=resolver_report.progress.total or None,
             )
 
-    def handle_success(self, main_task_id: TaskID, operation_display_type: str) -> None:
+    def handle_success(
+        self,
+        main_task_id: TaskID,
+        operation_display_type: str,
+        fqn_to_task_id: dict[str, TaskID],
+        shard_task_id: TaskID | None,
+    ):
+        if shard_task_id is not None and self.shard_progress is not None:
+            self.shard_progress.update(shard_task_id, advance=1)
         if self.enclosure_panel is not None:
-            self.enclosure_panel.title = "chalk ■"
+            if self._shard_id == self._num_computers:
+                self.enclosure_panel.title = "chalk ■"
         if self.main_progress is not None:
-            self.main_progress.update(main_task_id, description=f"{operation_display_type} completed", completed=1)
+            if self._shard_id == self._num_computers:
+                self.main_progress.update(main_task_id, description=f"{operation_display_type} completed", completed=1)
+        if self.resolver_progress is not None:
+            for task_id in fqn_to_task_id.values():
+                self.resolver_progress.reset(task_id)
 
     def _get_failing_resolver(self, batch_report: BatchReport) -> Optional[str]:
         """
@@ -241,9 +305,6 @@ class ProgressService:
         }
 
         metadata = batch_report.operation_metadata or {}
-        if not isinstance(metadata, Mapping):
-            metadata = {}
-
         metadata_tuples = [(key_to_display_key.get(k, k), v) for k, v in metadata.items()]
         failing_resolver = self._get_failing_resolver(batch_report)
 
@@ -270,73 +331,98 @@ class ProgressService:
 
     def handle_explainer_text_if_erring(self):
         if self.explainer_text:
-            # FIXME -- why are we modifying a private variable?
             assert self.caller_method is not None
-            self.explainer_text._text = [  # pyright: ignore[reportPrivateUsage]
+            self.explainer_text.plain = (
                 f"The computation for this `DatasetRevision` has failed. {self.caller_method}() cannot be executed."
-            ]
+            )
             self.explainer_text.style = Style(color="red")
 
     def handle_explainer_text_if_resolver_runtime_error(self):
         if self.explainer_text:
-            # FIXME -- why are we modifying a private variable?
             assert self.caller_method is not None
-            self.explainer_text._text = [  # pyright: ignore[reportPrivateUsage]
-                f"The computation for this `DatasetRevision` has failed due to a resolver runtime error. {self.caller_method}() can still be executed."
-            ]
+            self.explainer_text.plain = f"The computation for this `DatasetRevision` has failed due to a resolver runtime error. {self.caller_method}() can still be executed."
             self.explainer_text.style = Style(color="yellow")
 
-    def poll_report(self, computer_id: int, timeout: float | timedelta | None) -> Iterable[BatchReport]:
+    def poll_report(
+        self,
+        timeout: float | timedelta | None,
+        completion_timeout: float | timedelta | None,
+    ) -> Iterable[BatchReport]:
+        if completion_timeout is None:
+            completion_deadline = datetime.max
+        else:
+            if not isinstance(completion_timeout, timedelta):
+                completion_timeout = timedelta(seconds=completion_timeout)
+            completion_deadline = datetime.now() + completion_timeout
+
         if timeout is None:
-            expiration = None
+            must_receive_next_report_by = datetime.max
         else:
             if not isinstance(timeout, timedelta):
                 timeout = timedelta(seconds=timeout)
-            expiration = datetime.now() + timeout
+            must_receive_next_report_by = datetime.now() + timeout
 
-        while expiration is None or datetime.now() < expiration:
+        while datetime.now() < must_receive_next_report_by and datetime.now() < completion_deadline:
             time.sleep(0.5)
-            batch_report = self.client.get_batch_report(self.operation_id, self._environment_id, computer_id)
-            if batch_report is not None:
-                self._operation_kind = batch_report.operation_kind
-                yield batch_report
-        else:
-            raise TimeoutError(f"Timed out waiting for status report of operation with ID {self.operation_id}")
 
-    def await_operation(
-        self, computer_id: int, must_fail_on_resolver_error: bool, timeout: float | timedelta | None
-    ) -> None:
+            for attempt in Retrying(
+                stop=stop_after_attempt(5),
+                wait=wait_exponential_jitter(),
+                reraise=True,
+                retry=retry_if_exception_message(match="504|dialing"),
+            ):
+                with attempt:
+                    batch_report = self.client.get_batch_report(self.operation_id, self._environment_id, self._shard_id)
+                    if batch_report is not None:
+                        if timeout is not None:
+                            must_receive_next_report_by = datetime.now() + timeout
+                        self._operation_kind = batch_report.operation_kind
+                        yield batch_report
+        else:
+            if datetime.now() >= completion_deadline:
+                raise TimeoutError(
+                    f"Timed out waiting ({completion_timeout}) for completion of operation with id '{self.operation_id}' (chalkpy=={chalk.__version__})"
+                )
+            else:
+                raise TimeoutError(
+                    f"Timed out waiting ({timeout}) for next status report for operation with id '{self.operation_id}' (chalkpy=={chalk.__version__})"
+                )
+
+    def await_operation(self, must_fail_on_resolver_error: bool, timeout: float | timedelta | None) -> None:
         fqn_to_task_id: dict[str, TaskID] = {}
         initial_description = "Starting query execution"
         if self.main_progress is not None:
             main_task_id = self.main_progress.add_task(description=initial_description, total=1)
-            main_task = self.main_progress.tasks[main_task_id]
+            if self.shard_progress is not None:
+                shard_task_id = self.shard_progress.add_task(description="Processing Shards", total=self._num_computers)
+            else:
+                shard_task_id = None
         else:
             main_task_id = None
-            main_task = None
-
-        with ChalkLive(self.enclosure_panel, auto_refresh=True):
-            for batch_report in self.poll_report(computer_id, timeout):
-                operation_display_type = _kind_to_display_type.get(batch_report.operation_kind, "Operation")
+            shard_task_id = None
+        context = ChalkLive(self.enclosure_panel, auto_refresh=True) if self.show_progress else nullcontext()
+        with context:
+            for batch_report in self.poll_report(timeout, completion_timeout=self.client.default_job_timeout):
 
                 # Update main progress text
-                if main_task is not None:
+                if main_task_id is not None:
                     assert self.main_progress is not None
                     assert main_task_id is not None
-                    if main_task.description == initial_description:
-                        description = (
-                            f"Executing {operation_display_type}"
-                            if self._operation_kind == BatchOpKind.RECOMPUTE
-                            else f"Executing {operation_display_type}"
-                        )
+                    if self.main_progress.tasks[main_task_id].description == initial_description:
+                        description = f"Executing {OFFLINE_QUERY}: "
                         self.main_progress.update(main_task_id, description=description)
-
                 self.handle_resolver_update(batch_report=batch_report, fqn_to_task_id=fqn_to_task_id)
-
                 if batch_report.status == BatchOpStatus.COMPLETED:
+                    self._shard_id += 1
                     if main_task_id is not None:
-                        self.handle_success(main_task_id=main_task_id, operation_display_type=operation_display_type)
-                    break
+                        self.handle_success(
+                            main_task_id=main_task_id,
+                            operation_display_type=OFFLINE_QUERY,
+                            fqn_to_task_id=fqn_to_task_id,
+                            shard_task_id=shard_task_id,
+                        )
+                    if self._shard_id == self._num_computers:
+                        break
 
                 if batch_report.status == BatchOpStatus.FAILED:
                     if main_task_id is not None:
@@ -345,14 +431,14 @@ class ProgressService:
                             self.handle_and_raise_error(
                                 batch_report=batch_report,
                                 main_task_id=main_task_id,
-                                operation_display_type=operation_display_type,
+                                operation_display_type=OFFLINE_QUERY,
                                 is_fatal=must_fail_on_resolver_error,
                             )
                         else:
                             self.handle_and_raise_error(
                                 batch_report=batch_report,
                                 main_task_id=main_task_id,
-                                operation_display_type=operation_display_type,
+                                operation_display_type=OFFLINE_QUERY,
                             )
                     if must_fail_on_resolver_error:
                         raise ChalkBaseException(
