@@ -4,12 +4,13 @@ from typing import List, Optional
 
 from typing_extensions import Self
 
+import chalk
 from chalk.client.models import ChalkError
 
 CHALK_TRACE_ID_KEY = "trace"
 _CHALK_TRACE_ID_KEY = CHALK_TRACE_ID_KEY  # for backcompat
-_CHALK_TRACE_ID_MESSAGE = "Please contact Chalk Support with the Trace ID"
-_CHALK_TRACE_ID_MISSING = "<trace ID unexpectedly missing>"
+_CHALK_TRACE_ID_MESSAGE = "Trace ID:"
+_CHALK_TRACE_ID_MISSING = "<trace ID missing>"
 
 
 class ChalkBaseException(Exception):
@@ -41,7 +42,7 @@ class ChalkBaseException(Exception):
     @property
     def message(self) -> str:
         """A message describing the specific type of exception raised."""
-        return "Failed to execute Chalk operation"
+        return f"Failed to execute Chalk operation (chalkpy=={chalk.__version__})"
 
     @property
     def full_message(self) -> str:
@@ -53,12 +54,6 @@ class ChalkBaseException(Exception):
         """
         msg = self.message
 
-        if self.detail:
-            msg += f"\n\nAdditional details: \n\t{self.detail}"
-
-        if self.trace_id and _CHALK_TRACE_ID_MESSAGE not in msg:
-            msg += f"\n\n{_CHALK_TRACE_ID_MESSAGE}: \n\t{self.trace_id}"
-
         if self.errors:
             for e in self.errors[:10]:
                 if e.display_primary_key:
@@ -70,6 +65,13 @@ class ChalkBaseException(Exception):
                 if e.exception:
                     msg += f"\n{e.exception.stacktrace}"
                     msg += f"\n{e.exception.message}"
+
+        if self.detail:
+            msg += f"\n\nAdditional details: \n\t{self.detail}"
+
+        if self.trace_id and _CHALK_TRACE_ID_MESSAGE not in msg:
+            msg += f"\n\n{_CHALK_TRACE_ID_MESSAGE}: \n\t{self.trace_id}"
+
         return msg
 
 
