@@ -9,28 +9,44 @@ from chalk._lsp.error_builder import get_resolver_error_builder
 from chalk.features._embedding.cohere import CohereProvider
 from chalk.features._embedding.embedding_provider import EmbeddingProvider
 from chalk.features._embedding.openai import OpenAIProvider
+from chalk.features._embedding.sentence_transformer import SentenceTransformerProvider
+from chalk.features._embedding.vertexai import VertexAIProvider
 from chalk.features.dataframe import DataFrame
 from chalk.features.feature_field import Feature
 from chalk.features.feature_set import Features
 from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
 from chalk.features.resolver import RESOLVER_REGISTRY, OnlineResolver
+from chalk.features.underscore import Underscore, UnderscoreFunction
 from chalk.serialization.parsed_annotation import ParsedAnnotation
 from chalk.utils.collections import ensure_tuple
 
+SUPPORTED_LOCAL_MODELS = {
+    "all-MiniLM-L6-v2",  # https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2
+    "sample-bert",  # For internal Chalk use
+}
 
-def _get_provider(provider: str, model: str) -> EmbeddingProvider:
+
+def _get_provider(provider: str, model: str, dimensions: int | None = None) -> EmbeddingProvider:
     """Factory function to get an EmbeddingProvider"""
     if provider == "openai":
-        return OpenAIProvider(model)
+        return OpenAIProvider(model, dimensions)
     elif provider == "cohere":
-        return CohereProvider(model)
-    raise ValueError(f"Unsupported embedding provider: {provider}. The supported providers are ['openai', 'cohere'].")
+        return CohereProvider(model, dimensions)
+    elif provider == "vertexai":
+        return VertexAIProvider(model, dimensions)
+    elif provider == "sentence-transformers":
+        return SentenceTransformerProvider(model, dimensions)
+    raise ValueError(
+        f"Unsupported embedding provider: {provider}. The supported providers are ['openai', 'cohere', 'vertexai']."
+    )
 
 
 def embed(
     input: Callable[[], Any],
-    provider: str,
-    model: str,
+    model: str | None = None,
+    model_path: str | None = None,
+    tokenizer_path: str | None = None,
+    provider: str | None = None,
     name: str | None = None,
     owner: str | None = None,
     tags: list[str] | None = None,
@@ -40,6 +56,8 @@ def embed(
     default_version: int = 1,
     etl_offline_to_online: bool | None = None,
     offline_ttl: ellipsis | str | timedelta | None = ...,
+    default: Any = None,
+    dimensions: int | None = None,
 ) -> Any:
     """Specify an embedding feature.
 
@@ -53,6 +71,13 @@ def embed(
         The AI provider to use for the embedding.
     model
         The model to generate the embedding.
+    model_path
+        The path to the embedding model file: currently, `onnx` models are supported.
+    tokenizer_path
+        The path to the tokenizer for the model file. This is only used running an embedding model locally.
+    dimensions
+        The dimensionality of the embedding. If not specified, the default dimensionality will be used.
+        Supported dimensionality varies based on the model.
     owner
         You may also specify which person or group is responsible for a feature.
         The owner tag will be available in Chalk's web portal.
@@ -136,6 +161,9 @@ def embed(
         You can also specify the description directly with this parameter.
         Read more at https://docs.chalk.ai/docs/feature-discovery#description
     offline_ttl
+    default
+        The default value to use for this embedding feature when no value is available.
+        This can be useful when you want to provide a fallback embedding value.
 
     Examples
     --------
@@ -150,7 +178,26 @@ def embed(
     ...         model="text-embedding-ada-002",
     ...     )
     """
-    embedding_provider = _get_provider(provider, model)
+    if provider is None:
+        if not isinstance(input, Underscore):
+            raise TypeError(
+                "When using `model_path` and `provider`, the `input` must be an underscore not a lambda expression: `_.content`."
+            )
+
+        if model in SUPPORTED_LOCAL_MODELS:
+            return UnderscoreFunction("onnx_run_embedding", input, model_name=model)
+
+        if model_path is not None and tokenizer_path is not None:
+            return UnderscoreFunction("onnx_run_embedding", input, model_path=model_path, tokenizer_path=tokenizer_path)
+
+        raise ValueError(
+            f"If `provider` is None, you must specify either: both a `model_path` and `tokenizer_path`, or one of the supported local embedding models ({', '.join(list(SUPPORTED_LOCAL_MODELS))})."
+        )
+
+    if model is None:
+        raise ValueError("If provider is set, then model must also be set")
+
+    embedding_provider = _get_provider(provider, model, dimensions)
     # Manually set the dimensions of the Vector when using embedding
     typ = ParsedAnnotation(underlying=embedding_provider.get_vector_class())
     output_feature = Feature(
@@ -164,6 +211,7 @@ def embed(
         default_version=default_version,
         etl_offline_to_online=etl_offline_to_online,
         offline_ttl=offline_ttl,
+        default=default,
     )
     previous_hook = output_feature.hook
 
@@ -199,8 +247,8 @@ def embed(
                     )
                 )
 
-            def fn(raw_inputs: DataFrame):
-                # We need to build the input table by combinding the literals with the features
+            async def fn(raw_inputs: DataFrame):
+                # We need to build the input table by combining the literals with the features
                 input_arrays: list[pa.Array | pa.ChunkedArray] = []
                 raw_input_table = raw_inputs.to_pyarrow()
                 assert features.__chalk_primary__ is not None
@@ -213,10 +261,15 @@ def embed(
                 unified_inputs_table = pa.Table.from_arrays(
                     input_arrays, names=[f"col_{i}" for i in range(len(input_arrays))]
                 )
-                embeddings = embedding_provider.generate_embedding(unified_inputs_table)
-                return pa.Table.from_arrays(
-                    [embeddings, pkeys], [output_feature.root_fqn, features.__chalk_primary__.root_fqn]
-                )
+                i = 0
+                async for embeddings in embedding_provider.async_generate_embedding(unified_inputs_table):
+                    t = pa.Table.from_arrays(
+                        [embeddings, pkeys.slice(i, len(embeddings))],
+                        [output_feature.root_fqn, features.__chalk_primary__.root_fqn],
+                    )
+                    i += len(embeddings)
+
+                    yield t
 
             return OnlineResolver(
                 function_definition="",
@@ -239,8 +292,15 @@ def embed(
                 is_sql_file_resolver=False,
                 source_line=None,
                 lsp_builder=get_resolver_error_builder(fn),
-                underscore=None,
                 parse=None,
+                resource_hint=None,
+                static=False,
+                total=False,
+                autogenerated=False,
+                unique_on=None,
+                partitioned_by=None,
+                data_lineage=None,
+                sql_settings=None,
             )
 
         RESOLVER_REGISTRY.add_to_deferred_registry(resolver_factory, override=False)
