@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import collections.abc
 import dataclasses
 import decimal
 import enum
@@ -7,18 +8,21 @@ import ipaddress
 import typing
 import uuid
 from datetime import date, datetime, time, timedelta
-from typing import TYPE_CHECKING, Any, Dict, FrozenSet, List, Set, Tuple, Type, cast
+from typing import TYPE_CHECKING, Any, Dict, FrozenSet, List, Mapping, Set, Tuple, Type, cast
 
 import attrs
 import pyarrow as pa
 from typing_extensions import Annotated, TypeGuard, get_args, get_origin, is_typeddict
 
+from chalk.features._encoding.http import HttpResponse, get_http_response_as_pyarrow
 from chalk.features._encoding.primitive import ChalkStructType, TPrimitive
 from chalk.features.feature_set import Features
 from chalk.utils.cached_type_hints import cached_get_type_hints
 from chalk.utils.collections import is_namedtuple, is_optional, unwrap_optional_and_annotated_if_needed
 from chalk.utils.enum import get_enum_value_type
+from chalk.utils.json import JSON, is_pyarrow_json_type
 from chalk.utils.missing_dependency import missing_dependency_exception
+from chalk.utils.pl_helpers import is_new_polars
 from chalk.utils.pydanticutil.pydantic_compat import is_pydantic_basemodel
 
 if TYPE_CHECKING:
@@ -28,7 +32,7 @@ if TYPE_CHECKING:
 
 try:
     import numpy as np
-except:
+except ImportError:
     np = None
 
 __all__ = ["pyarrow_to_primitive", "pyarrow_to_polars", "rich_to_pyarrow"]
@@ -45,13 +49,22 @@ def _is_chalk_dataframe(typ: Any) -> TypeGuard[type[DataFrame]]:
     return isinstance(unwrapped, type) and issubclass(unwrapped, DataFrame)
 
 
-def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa.DataType:
+def _is_chalk_http_response(typ: Any) -> TypeGuard[type[HttpResponse]]:
+    from chalk.features._encoding.http import HttpResponse
+
+    return isinstance(typ, type) and issubclass(typ, HttpResponse)
+
+
+def rich_to_pyarrow(
+    python_type: Type, name: str, in_struct: bool = False, respect_nullability: bool = False
+) -> pa.DataType:
     """
     Recursively convert a python type to a PyArrow dtype.
     :param python_type:
     :param name:
     :param in_struct: Whether this function is being called recursively for a member of a struct type.
-    Certain datatypes need to be handled differently if nested inside a struct.
+                      Certain datatypes need to be handled differently if nested inside a struct.
+    :param respect_nullability: Whether this function should update pa.field() with correct nullability.
     """
     # Polars seems to allow optional for any dtype, so we ignore it when computing dtypes
     python_type = unwrap_optional_and_annotated_if_needed(python_type)
@@ -65,7 +78,7 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
                 raise TypeError(
                     "Annotated types must contain the underlying type as the first argument -- e.g. Annotated[int, 'annotation']."
                 )
-            return rich_to_pyarrow(args[0], name, in_struct=in_struct)
+            return rich_to_pyarrow(args[0], name, in_struct=in_struct, respect_nullability=respect_nullability)
         if origin in (list, List, set, Set, frozenset, FrozenSet):
             typ_name = origin.__name__
             if len(args) == 0:
@@ -76,8 +89,10 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
                 raise TypeError(
                     f"{typ_name} annotations should only take one argument -- e.g. {typ_name}[int]. Instead, got {typ_name}[{', '.join(args)}]."
                 )
-            arg = args[0]
-            return pa.large_list(rich_to_pyarrow(arg, name=f"{name}[]", in_struct=in_struct))
+            arg = cast("tuple[Any, ...]", args)[0]
+            return pa.large_list(
+                rich_to_pyarrow(arg, name=f"{name}[]", in_struct=in_struct, respect_nullability=respect_nullability)
+            )
         if origin in (tuple, Tuple):
             if len(args) == 0:
                 raise TypeError(
@@ -86,7 +101,9 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
             if len(args) == 2 and args[1] is ...:
                 # Treat a variable-sized, homogenous tuple like a list
                 arg = args[0]
-                return pa.large_list(rich_to_pyarrow(arg, name=f"{name}[]", in_struct=in_struct))
+                return pa.large_list(
+                    rich_to_pyarrow(arg, name=f"{name}[]", in_struct=in_struct, respect_nullability=respect_nullability)
+                )
             raise TypeError(
                 (
                     "Tuple features must have a fixed type and be variable-length tuples (e.g. `Tuple[int, ...]`). "
@@ -94,7 +111,8 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
                 )
             )
 
-        if origin in (dict, Dict):
+        if origin in (dict, Dict, Mapping, collections.abc.Mapping):
+            # pydantic changes typing.Mapping into collections.abc.Mapping
             if len(args) != 2:
                 raise TypeError(
                     "Dict features must be annotated with the type of the key and value -- e.g. `Dict[str, int]`."
@@ -107,21 +125,36 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
                 raise TypeError(f"Dict keys must be annotated with a `type`, found `{key_type}`")
 
             try:
-                pa_value_type = rich_to_pyarrow(value_type, "value")
+                pa_value_type = rich_to_pyarrow(value_type, "value", respect_nullability=respect_nullability)
             except TypeError as e:
                 raise TypeError(f"Failed to determine type of dict value from the annotation `{value_type}`") from e
             else:
                 value_field = pa.field(name="value", type=pa_value_type, nullable=is_optional(value_type))
 
-                return pa.map_(key_type=rich_to_pyarrow(key_type, "key"), item_type=value_field)
+                return pa.map_(
+                    key_type=rich_to_pyarrow(key_type, "key", respect_nullability=respect_nullability),
+                    item_type=value_field,
+                )
+
+        if _is_chalk_http_response(origin):
+            return get_http_response_as_pyarrow(python_type)
 
         raise TypeError(f"Unsupported varardic type annotation: {origin}")
     else:
-        if python_type in (list, List, set, Set, frozenset, FrozenSet, tuple, Tuple):
+        if python_type in (list, List, set, Set, frozenset, FrozenSet):
+            typ_name = python_type.__name__
             raise TypeError(
                 (
-                    f"Unable to determine the PyArrow type for field '{name}' with type `{python_type}`. "
-                    "List types must be parameterized by their element type, e.g. list[str] or list[int]. "
+                    f"Unable to determine the PyArrow type for field '{name}' with type `{typ_name}`. "
+                    f"{typ_name} features must be parameterized by their element type, e.g. {typ_name}[str]. "
+                )
+            )
+        elif python_type in (tuple, Tuple, dict, Dict, Mapping, collections.abc.Mapping):
+            typ_name = python_type.__name__
+            raise TypeError(
+                (
+                    f"Unable to determine the PyArrow type for field '{name}' with type `{typ_name}`. "
+                    f"{typ_name} features must be parameterized by their element type, e.g. {typ_name}[str, int]. "
                 )
             )
 
@@ -129,17 +162,20 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
 
     if isinstance(python_type, Windowed):
         python_type = python_type.kind
-    if not isinstance(python_type, type):
+    if not isinstance(python_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
         raise TypeError(f"Type annotations must be a type. Instead, got {python_type}.")
     if issubclass(python_type, enum.Enum):
         # For enums, require all members to have the same type
-        return rich_to_pyarrow(get_enum_value_type(python_type), name, in_struct=in_struct)
+        return rich_to_pyarrow(
+            get_enum_value_type(python_type), name, in_struct=in_struct, respect_nullability=respect_nullability
+        )
     # First, handle the recursive types -- list, tuple, typeddict, namedtuple, dataclass, pydantic model
     if _is_features_cls(python_type):
         annotations = cached_get_type_hints(python_type)
         fields: List[pa.Field] = []
         for field_name, type_annotation in annotations.items():
-            if _is_features_cls(type_annotation) or _is_chalk_dataframe(type_annotation):
+            type_annotation_unwrapped = unwrap_optional_and_annotated_if_needed(type_annotation)
+            if _is_features_cls(type_annotation_unwrapped) or _is_chalk_dataframe(type_annotation_unwrapped):
                 continue
             features_cls = unwrap_optional_and_annotated_if_needed(python_type)
             feature_wrapper = getattr(features_cls, field_name)
@@ -161,8 +197,16 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
             if field_name.startswith("__"):
                 # Skip all dunders, like __slots__
                 continue
-            underlying_dtype = rich_to_pyarrow(type_annotation, name=f"{name}.{field_name}", in_struct=True)
-            fields.append(pa.field(field_name, underlying_dtype))
+            underlying_dtype = rich_to_pyarrow(
+                type_annotation, name=f"{name}.{field_name}", in_struct=True, respect_nullability=respect_nullability
+            )
+            fields.append(
+                pa.field(
+                    field_name,
+                    underlying_dtype,
+                    not respect_nullability or type(None) in get_args(type_annotation),
+                )
+            )
         return pa.struct(fields)
     if _is_chalk_dataframe(python_type):
         if python_type.__pydantic_model__ is None:
@@ -172,7 +216,13 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
                 pa.struct([pa.field(col.root_fqn, col.converter.pyarrow_dtype) for col in python_type.columns])
             )
         else:
-            return pa.large_list(rich_to_pyarrow(python_type=python_type.__pydantic_model__, name="item"))
+            return pa.large_list(
+                rich_to_pyarrow(
+                    python_type=python_type.__pydantic_model__, name="item", respect_nullability=respect_nullability
+                )
+            )
+    if issubclass(python_type, type(None)):
+        return pa.null()
     if issubclass(python_type, str):
         return pa.large_utf8()
     if issubclass(python_type, bool):
@@ -217,33 +267,35 @@ def rich_to_pyarrow(python_type: Type, name: str, in_struct: bool = False) -> pa
 def pyarrow_to_primitive(pyarrow_typ: pa.DataType, name: str) -> Type[TPrimitive]:
     if pa.types.is_null(pyarrow_typ):
         return type(None)
-    if pa.types.is_boolean(pyarrow_typ):
+    elif pa.types.is_boolean(pyarrow_typ):
         return bool
-    if pa.types.is_unsigned_integer(pyarrow_typ) or pa.types.is_signed_integer(pyarrow_typ):
+    elif pa.types.is_unsigned_integer(pyarrow_typ) or pa.types.is_signed_integer(pyarrow_typ):
         return int
-    if pa.types.is_floating(pyarrow_typ):
+    elif pa.types.is_floating(pyarrow_typ):
         return float
-    if pa.types.is_time(pyarrow_typ):
+    elif pa.types.is_time(pyarrow_typ):
         return time
-    if pa.types.is_timestamp(pyarrow_typ):
+    elif pa.types.is_timestamp(pyarrow_typ):
         return datetime
-    if pa.types.is_date(pyarrow_typ):
+    elif pa.types.is_date(pyarrow_typ):
         return date
-    if pa.types.is_duration(pyarrow_typ):
+    elif pa.types.is_duration(pyarrow_typ):
         return timedelta
-    if (
+    elif (
         pa.types.is_binary(pyarrow_typ)
         or pa.types.is_fixed_size_binary(pyarrow_typ)
         or pa.types.is_large_binary(pyarrow_typ)
     ):
         return bytes
-    if pa.types.is_string(pyarrow_typ) or pa.types.is_large_string(pyarrow_typ):
+    elif pa.types.is_string(pyarrow_typ) or pa.types.is_large_string(pyarrow_typ):
         return str
-    if pa.types.is_list(pyarrow_typ) or pa.types.is_large_list(pyarrow_typ) or pa.types.is_fixed_size_list(pyarrow_typ):
+    elif (
+        pa.types.is_list(pyarrow_typ) or pa.types.is_large_list(pyarrow_typ) or pa.types.is_fixed_size_list(pyarrow_typ)
+    ):
         assert isinstance(pyarrow_typ, (pa.ListType, pa.LargeListType, pa.FixedSizeListType))
         underlying = pyarrow_typ.value_type
         return List[pyarrow_to_primitive(underlying, name=f"{name}[]")]
-    if pa.types.is_struct(pyarrow_typ):
+    elif pa.types.is_struct(pyarrow_typ):
         assert isinstance(pyarrow_typ, pa.StructType)
 
         annotations: Dict[str, Type[TPrimitive]] = {}
@@ -256,7 +308,7 @@ def pyarrow_to_primitive(pyarrow_typ: pa.DataType, name: str) -> Type[TPrimitive
             annotations[sub_name] = field_type
         struct_type = ChalkStructType(f"__chalk_struct__{name}", (object,), annotations)
         return cast("type[ChalkStructType]", struct_type)
-    if pa.types.is_map(pyarrow_typ):
+    elif pa.types.is_map(pyarrow_typ):
         assert isinstance(pyarrow_typ, pa.MapType)
         value_typ = pyarrow_to_primitive(pyarrow_typ.item_type, name=f"{name}::value")
         if pyarrow_typ.item_field.nullable:
@@ -265,6 +317,8 @@ def pyarrow_to_primitive(pyarrow_typ: pa.DataType, name: str) -> Type[TPrimitive
             pyarrow_to_primitive(pyarrow_typ.key_type, name=f"{name}::key"),
             value_typ,
         ]
+    elif is_pyarrow_json_type(pyarrow_typ):
+        return cast(Type[TPrimitive], JSON)
     raise TypeError(f"Unsupported PyArrow type '{pyarrow_typ}' for field '{name}'.")
 
 
@@ -278,7 +332,9 @@ def is_map_in_dtype_tree(dtype: pa.DataType) -> bool:
     return False
 
 
-def pyarrow_to_polars(pa_type: pa.DataType, name: str | None = None) -> pl.PolarsDataType:
+def pyarrow_to_polars(
+    pa_type: pa.DataType, name: str | None = None, use_fixed_size_list: bool = False
+) -> pl.PolarsDataType:
     """Convert a PyArrow data type into a Polars DataType
 
     Args:
@@ -289,8 +345,12 @@ def pyarrow_to_polars(pa_type: pa.DataType, name: str | None = None) -> pl.Polar
         import polars as pl
     except ImportError:
         raise missing_dependency_exception("chalkpy[runtime]")
+    if isinstance(pa_type, pa.ExtensionType):
+        pa_type = pa_type.storage_type
     if pa.types.is_null(pa_type):
         return pl.Null()
+    if is_pyarrow_json_type(pa_type):
+        return pl.Utf8()
     if pa.types.is_boolean(pa_type):
         return pl.Boolean()
     if pa.types.is_int8(pa_type):
@@ -331,10 +391,17 @@ def pyarrow_to_polars(pa_type: pa.DataType, name: str | None = None) -> pl.Polar
         return pl.Binary()
     if pa.types.is_string(pa_type) or pa.types.is_large_string(pa_type):
         return pl.Utf8()
-    if pa.types.is_list(pa_type) or pa.types.is_large_list(pa_type) or pa.types.is_fixed_size_list(pa_type):
+    if pa.types.is_list(pa_type) or pa.types.is_large_list(pa_type):
         assert isinstance(pa_type, (pa.ListType, pa.LargeListType, pa.FixedSizeListType))
         underlying = pa_type.value_type
         return pl.List(pyarrow_to_polars(underlying, name=f"{name}[]"))
+    if pa.types.is_fixed_size_list(pa_type):
+        underlying = pa_type.value_type
+        if is_new_polars and use_fixed_size_list:
+            # pl.Array is only available in polars >=0.18
+            return pl.Array(inner=pyarrow_to_polars(underlying, name=f"{name}[]"), width=pa_type.list_size)
+        else:
+            return pl.List(pyarrow_to_polars(underlying, name=f"{name}[]"))
     if pa.types.is_struct(pa_type):
         assert isinstance(pa_type, pa.StructType)
         schema = pa.schema(pa_type)
