@@ -1,8 +1,10 @@
 from __future__ import annotations
 
-from datetime import timedelta
+import contextvars
 import inspect
 import textwrap
+from collections import defaultdict
+from datetime import timedelta
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -15,46 +17,56 @@ from typing import (
     Tuple,
     Type,
     cast,
+    Set,
+    Protocol,
+    Mapping,
 )
 
-from typing_extensions import Final, TypeGuard
-from chalk.utils import notebook
+from typing_extensions import TypeGuard
+
 from chalk._lsp._class_finder import get_class_ast
 from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
-from chalk.features.live_updates import register_live_updates_if_in_notebook
+from chalk.utils import notebook
 from chalk.utils.collections import ensure_tuple, get_unique_item
+from chalk.utils.source_parsing import should_skip_source_code_parsing
 
 if TYPE_CHECKING:
     from chalk.features.feature_field import Feature
     from chalk._lsp.error_builder import FeatureClassErrorBuilder
+    from chalk.features.underscore_features import NamedUnderscoreExpr
 
-__all__ = ["Features", "FeaturesMeta", "FeatureSetBase", "is_features_cls"]
+__all__ = ["Features", "FeaturesMeta", "FeatureSetBase", "is_features_cls", "is_feature_set_class"]
 
 
 class ClassSource:
-
     __slots__ = ("filename", "class_source", "tree", "dedent_source")
 
     def __init__(self, c: Any):
+        self.filename = None
+        self.class_source = None
+        self.tree = None
         super().__init__()
-        try:
-            self.filename: Final = inspect.getfile(c)
-        except:
-            self.filename = None
+        if not should_skip_source_code_parsing():
+            try:
+                self.filename = inspect.getfile(c)
+            except:
+                self.filename = None
 
-        try:
-            if notebook.is_defined_in_module(c):
-                self.class_source: Final = inspect.getsource(c)
-            # TODO (rkargon) Try to get source from notebook cells if possible
-            else:
+            try:
+                if notebook.is_defined_in_module(c):
+                    self.class_source = inspect.getsource(c)
+                # TODO (rkargon) Try to get source from notebook cells if possible
+                else:
+                    self.class_source = None
+            except:
                 self.class_source = None
-        except:
-            self.class_source = None
-        self.dedent_source: Final = self.class_source and textwrap.dedent(self.class_source)
-        try:
-            self.tree: Final = get_class_ast(c)
-        except:
-            self.tree = None
+            self.dedent_source = self.class_source and textwrap.dedent(self.class_source)
+            try:
+                self.tree = get_class_ast(c)
+            except:
+                self.tree = None
+
+
 class FeaturesMeta(type):
     """Metaclass for classes decorated with ``@features``.
 
@@ -65,10 +77,14 @@ class FeaturesMeta(type):
     2.  ``Features[Feature1, Feature2]`` annotations to return subclasses of Features, so they can be used as proper type annotations.
     """
 
-    features: List[Feature]  # pyright: ignore[reportUninitializedInstanceVariable]
+    _features_lazy: Callable[[], List[Feature]]  # pyright: ignore[reportUninitializedInstanceVariable]
+
+    @property
+    def features(self) -> List[Feature]:
+        return self._features_lazy()
 
     def __subclasscheck__(self, subcls: type) -> bool:
-        if not isinstance(subcls, type):
+        if not isinstance(subcls, type):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise TypeError(f"{subcls} is not a type")
         if hasattr(subcls, "__is_features__"):
             return getattr(subcls, "__is_features__")
@@ -95,30 +111,33 @@ class FeaturesMeta(type):
                 for f in item.features
                 if not f.is_autogenerated and not f.is_has_one and not f.is_has_many and not f.is_windowed
             ]
-        item = ensure_tuple(item)
-        item = tuple(unwrap_feature(x) if isinstance(x, FeatureWrapper) else x for x in item)
-        for x in item:
-            if isinstance(x, str):
-                raise TypeError(
-                    (
-                        f'String features like {cls.__name__}["{x}"] are unsupported. '
-                        f'Instead, replace with "{cls.__name__}[{x}]"'
+
+        def f():
+            item_ = ensure_tuple(item)
+            item_ = tuple(unwrap_feature(x) if isinstance(x, FeatureWrapper) else x for x in item_)
+            for x in item_:
+                if isinstance(x, str):
+                    raise TypeError(
+                        (
+                            f'String features like {cls.__name__}["{x}"] are unsupported. '
+                            f'Instead, replace with "{cls.__name__}[{x}]"'
+                        )
                     )
-                )
 
-            if isinstance(x, Feature) or (isinstance(x, type) and issubclass(x, DataFrame)):
-                continue
+                if isinstance(x, Feature) or (isinstance(x, type) and issubclass(x, DataFrame)):  # pyright: ignore[reportUnnecessaryIsInstance]
+                    continue
 
-            raise TypeError(f"Invalid feature {x} of type '{type(x).__name__}'")
-        cls = cast(Type[Features], cls)
+                raise TypeError(f"Invalid feature {x} of type '{type(x).__name__}'")
+            cls_ = cast(Type[Features], cls)
 
-        new_features = list([*getattr(cls, "features", ()), *item])
-        if len(new_features) == 0:
-            raise ValueError("Cannot create empty Features[] class")
+            new_features = list([*getattr(cls_, "features", ()), *item_])
+            if len(new_features) == 0:
+                raise ValueError("Cannot create empty Features[] class")
+            return new_features
 
-        class SubFeatures(cls):
+        class SubFeatures(cls):  # pyright: ignore
             __is_exploded__ = is_exploded
-            features = new_features
+            _features_lazy = f
 
         return SubFeatures
 
@@ -149,7 +168,7 @@ class FeaturesImpl(metaclass=FeaturesMeta):
 
     This class is never instantiated or directly inherited. However, classes
     annotated with @features can be thought of as inheriting from this class.
-    It can be used with ``isinstance`` and ``issubclass`` checks, as well as for
+    It can be used with `isinstance` and `issubclass` checks, as well as for
     typing.
     """
 
@@ -169,6 +188,10 @@ class FeaturesImpl(metaclass=FeaturesMeta):
     __chalk_source_info__: "ClassVar[ClassSource]"
     __chalk_is_loaded_from_notebook__: ClassVar[bool] = False
     """True if loaded on a branch server from a pickle"""
+    __chalk_materialized_windows__: ClassVar[List[Feature]]
+    __chalk_group_by_materialized_windows__: ClassVar[List[Feature]]
+    __chalk_expression_windows__: ClassVar[List[Feature]]
+    __chalk_notebook_feature_expressions__: ClassVar[Dict[str, NamedUnderscoreExpr]]
 
     __is_exploded__: ClassVar[bool] = False
     __is_features__: ClassVar[bool] = True
@@ -188,20 +211,22 @@ class FeaturesImpl(metaclass=FeaturesMeta):
         raise NotImplementedError
 
 
-# Hack to get VSCode/Pylance/Pyright to type Features as Type[FeatureImpl]
-# but IntelliJ to type it as Type[Any]
-# Vscode can parse through literal dicts; IntelliJ can't
 if TYPE_CHECKING:
 
     class FeaturesProtocol(FeaturesImpl):
-        def __class_getitem__(cls, item: Any) -> "FeaturesProtocol":
-            ...
+        def __class_getitem__(cls, item: Any) -> "FeaturesProtocol": ...
 
-    _dummy_dict = {"0": FeaturesProtocol}
+    Features = FeaturesProtocol
 else:
-    _dummy_dict = {"0": FeaturesImpl}
+    Features = FeaturesImpl
 
-Features = _dummy_dict["0"]
+
+class FeatureRegistryProtocol(Protocol):
+    def get_feature_sets(self) -> Mapping[str, Type[Features]]: ...
+
+    def add_feature_set(self, updated_class: Type[Features]): ...
+
+    def get_singletons(self) -> Mapping[str, Type[Features]]: ...
 
 
 class FeatureSetBase:
@@ -213,6 +238,12 @@ class FeatureSetBase:
     __chalk_singletons__: ClassVar[Dict[str, Type[Features]]] = {}
     """Mapping of fqn to singleton Features cls. These are also included in `.registry`."""
 
+    __chalk_notebook_defined_feature_fields__: ClassVar[Dict[str, Set[str]]] = defaultdict(set)
+    """
+    (Mapping of namespace to a set of unversioned attribute names).
+    This tracks which feature fields have been added to a feature class in a notebook, after it's been defined.
+    """
+
     hook: ClassVar[Optional[Callable[[Type[Features]], None]]] = None
 
     def __init__(self) -> None:
@@ -220,7 +251,59 @@ class FeatureSetBase:
         raise RuntimeError("FeatureSetBase should never be instantiated")
 
 
-register_live_updates_if_in_notebook(FeatureSetBase)
+class FeatureSetBaseWrapper(FeatureRegistryProtocol):
+    """
+    Wrapper around the global FeatureSetBase
+    """
+
+    def __init__(self):
+        super().__init__()
+
+    def get_feature_sets(self) -> Mapping[str, Type[Features]]:
+        return FeatureSetBase.registry
+
+    def add_feature_set(self, updated_class: Type[Features]):
+        namespace = updated_class.namespace
+        previous_features_class = FeatureSetBase.registry.get(namespace)
+        if (
+            notebook.is_notebook()
+            and previous_features_class is not None
+            and notebook.is_defined_in_module(previous_features_class)
+        ):
+            # Not generating an LSP here because we're in a notebook anyway
+            # TODO: See if we can pretty-print lsp errors in notebooks, at which point we can generate one that points to the old feature class
+            raise ValueError(
+                f"Cannot re-define feature class '{previous_features_class.__name__}' in a notebook: it was previously defined in '{previous_features_class.__module__}'."
+            )
+
+        FeatureSetBase.registry[updated_class.__chalk_namespace__] = updated_class
+
+        if updated_class.__chalk_is_singleton__:
+            FeatureSetBase.__chalk_singletons__[updated_class.__chalk_namespace__] = updated_class
+
+        if FeatureSetBase.hook is not None:
+            try:
+                FeatureSetBase.hook(updated_class)
+            except:
+                # If we were unsuccessful, restore the previous features class, and call the hook again
+                if previous_features_class is not None:
+                    FeatureSetBase.registry[updated_class.__chalk_namespace__] = previous_features_class
+                    if updated_class.__chalk_is_singleton__:
+                        FeatureSetBase.__chalk_singletons__[updated_class.__chalk_namespace__] = previous_features_class
+                    FeatureSetBase.hook(previous_features_class)
+                # Now raise the original exception
+                raise
+
+        ...
+
+    def get_singletons(self) -> Mapping[str, Type[Features]]:
+        return FeatureSetBase.__chalk_singletons__
+
+
+_DEFAULT_FEATURE_REGISTRY: FeatureRegistryProtocol = FeatureSetBaseWrapper()
+CURRENT_FEATURE_REGISTRY: contextvars.ContextVar[FeatureRegistryProtocol] = contextvars.ContextVar(
+    "CURRENT_FEATURE_REGISTRY", default=_DEFAULT_FEATURE_REGISTRY
+)
 
 
 def is_features_cls(maybe_features: Any) -> TypeGuard[Type[Features]]:
