@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import copy
+import dataclasses
 import inspect
 from collections.abc import Iterable
 from typing import Any, Literal, TypeVar, Union, TYPE_CHECKING, Callable, overload
@@ -9,7 +10,6 @@ from chalk._lsp.error_builder import LSPErrorBuilder
 from chalk.features._chalkop import op, Aggregation
 from chalk.features.filter import Filter
 from chalk.serialization.parsed_annotation import ParsedAnnotation
-from chalk.streams import get_name_with_duration
 from chalk.utils.collections import ensure_tuple
 
 if TYPE_CHECKING:
@@ -93,6 +93,7 @@ class FeatureWrapper:
         frame = inspect.currentframe()
         if frame and frame.f_back and frame.f_back.f_globals["__name__"] in ("typing", "typing_extensions"):
             return 0
+        del frame
         return hash(self._chalk_get_underlying())
 
     def __gt__(self, other: object):
@@ -133,14 +134,14 @@ class FeatureWrapper:
                 return False
             return NotImplemented  # GT / LT doesn't really make sense otherwise
         if isinstance(other, type):
-            return False
+            return NotImplemented
         return Filter(self, op, other)
 
     def __ne__(self, other: object):
         return self._cmp("!=", other)
 
     def __eq__(self, other: object):
-        if hasattr(other, "__module__") and  other.__module__ in ("typing", "typing_extensions"):
+        if hasattr(other, "__module__") and other.__module__ in ("typing", "typing_extensions"):
             return False
         return self._cmp("==", other)
 
@@ -153,7 +154,7 @@ class FeatureWrapper:
         if isinstance(other, type):
             # The FeatureWrapper appears as a UnionType in a type annotation -- e.g.
             # def my_resolver(name: User.name | None = None, ...)
-            return Union[other, self]  # type: ignore
+            return Union[other, self]
         return self._cmp("or", other)
 
     def __str__(self):
@@ -181,13 +182,17 @@ class FeatureWrapper:
         if not isinstance(window, (str, int)):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise TypeError("Window duration must be a string or an int")
 
-        from chalk.features import FeatureSetBase, Feature
+        from chalk.features import Feature
+        from chalk.features.feature_set import CURRENT_FEATURE_REGISTRY
+        from chalk.streams import get_name_with_duration
+
         underlying_feature = self._chalk_get_underlying()
         if not isinstance(underlying_feature, Feature):
             raise TypeError(f"Cannot get the windowed feature for {underlying_feature}, which is not a feature.")
 
+        registry = CURRENT_FEATURE_REGISTRY.get().get_feature_sets()
         parent = (
-            FeatureSetBase.registry[underlying_feature.namespace]
+            registry[underlying_feature.namespace]
             if len(underlying_feature.path) == 0
             else FeatureWrapper(underlying_feature.path[-1].parent)
         )
@@ -214,8 +219,8 @@ class FeatureWrapper:
         def f():
             from chalk.features.feature_field import get_distance_feature_name, Feature
             from chalk.features.pseudofeatures import Distance
-            from chalk.features.feature_set import FeatureSetBase
-
+            from chalk.features.feature_set import CURRENT_FEATURE_REGISTRY
+            registry = CURRENT_FEATURE_REGISTRY.get()
             underlying = self._chalk_get_underlying()
             if not isinstance(underlying, Feature):
                 return underlying[item]  # pyright: ignore -- we want to pass through if the underlying isn't a feature
@@ -247,23 +252,27 @@ class FeatureWrapper:
                             op=filter.operation,
                             foreign_name=foreign_feature.name,
                         )
-                        x = next(
-                            f for f in FeatureSetBase.registry[foreign_feature.namespace].features if f.name == key)
+                        feature_fields = registry.get_feature_sets()[foreign_feature.namespace].features
+                        x = next(f for f in feature_fields if f.name == key)
                         dataframe_item[i] = x
                 tuple_item = tuple(dataframe_item)
 
-                f_copy = FeatureWrapper(copy.copy(underlying))
-                f_copy_underlying = f_copy._chalk_get_underlying()
-                assert isinstance(f_copy_underlying, Feature)
+                f_copy_underlying = copy.copy(underlying)
                 f_copy_underlying.typ = ParsedAnnotation(underlying=dataframe_typ[tuple_item])
-                if (
-                    hasattr(f_copy_underlying, "path")
-                    and len(f_copy_underlying.path) > 0
-                ):
-                    f_copy_underlying.path[-1].child = unwrap_feature(f_copy)
+                if len(f_copy_underlying.path) > 0:
+                    path_copy = list(f_copy_underlying.path)
+                    path_copy[-1] = dataclasses.replace(path_copy[-1], child=f_copy_underlying)
+                    f_copy_underlying.path = tuple(path_copy)
 
-                return f_copy
-            raise TypeError(f"Feature '{self}' does not support subscripting. Attempted to subscript with '{item}'")
+                return FeatureWrapper(f_copy_underlying)
+            item_features_maybe = []
+            if isinstance(item, Iterable):
+                for i in item:
+                    try:
+                        item_features_maybe.append(unwrap_feature(i).fqn)
+                    except:
+                        break
+            raise TypeError(f"Feature '{self}' of type {underlying.typ} does not support subscripting. Attempted to subscript with '{item_features_maybe or item}'")
 
         return FeatureWrapper(_MarkedUnderlyingFeature(f, ("__getitem__", item)))
 
@@ -281,6 +290,10 @@ class FeatureWrapper:
             if not isinstance(underlying, Feature):
                 return getattr(underlying, item)
             joined_class = underlying.joined_class
+            if joined_class is None:
+                underlying_df = underlying.typ.as_dataframe()
+                joined_class = underlying_df.references_feature_set if underlying_df is not None else None
+
             if joined_class is not None:
                 for f in joined_class.features:
                     assert isinstance(f, Feature), f"HasOne feature {f} does not inherit from FeaturesBase"
@@ -289,7 +302,8 @@ class FeatureWrapper:
 
                 assert underlying.features_cls is not None
                 underlying.features_cls.__chalk_error_builder__.invalid_attribute(
-                    joined_class.namespace,
+                    root_feature_str=joined_class.namespace,
+                    root_is_feature_class=True,
                     item=item,
                     candidates=[f.name for f in joined_class.features],
                     back=1,
@@ -299,10 +313,12 @@ class FeatureWrapper:
 
             assert underlying.features_cls is not None
             underlying.features_cls.__chalk_error_builder__.invalid_attribute(
-                underlying.features_cls.namespace,
+                root_feature_str=underlying.fqn,
+                root_is_feature_class=False,
                 item=item,
                 candidates=[],
                 back=1,
+                saved_frame=self,
             )
             assert False, "unreachable"
         LSPErrorBuilder.save_node(self, item)
@@ -360,6 +376,8 @@ def unwrap_feature(maybe_feature_wrapper: Any, raise_error: bool = True) -> Feat
     """
     from chalk.features.feature_field import Feature
 
+    if isinstance(maybe_feature_wrapper, str):
+        return Feature.from_root_fqn(maybe_feature_wrapper)
     if isinstance(maybe_feature_wrapper, FeatureWrapper):
         maybe_feature_wrapper = maybe_feature_wrapper._chalk_get_underlying()  # pyright: ignore[reportPrivateUsage]
     if isinstance(maybe_feature_wrapper, Feature):
