@@ -1,7 +1,7 @@
 from dataclasses import dataclass
 from datetime import datetime
 from enum import Enum
-from typing import Any, List, Optional
+from typing import Any, Dict, List, Optional
 
 import dataclasses_json
 
@@ -81,6 +81,9 @@ class UpsertScalarKindGQL:
     baseClasses: Optional[List[str]] = None  # Deprecated. Use the `dtype` instead
     hasEncoderAndDecoder: bool = False  # Deprecated. Use the `dtype` instead
     scalarKind: Optional[str] = None  # Deprecated. Use the `dtype` instead
+    isDeprecated: Optional[bool] = False
+    defaultValueJson: Optional[str] = None
+    hasExplicitDtype: Optional[bool] = False
 
 
 @dataclasses_json.dataclass_json
@@ -89,6 +92,23 @@ class UpsertFeatureTimeKindGQL:
     format: Optional[str] = None
 
 
+@dataclasses_json.dataclass_json
+@dataclass
+class UpsertWindowMaterializationGQL:
+    namespace: str
+    groupBy: List[UpsertFeatureReferenceGQL]
+    bucketDuration: float
+    aggregation: str
+    aggregateOn: Optional[UpsertFeatureReferenceGQL]
+    dtype: Optional[str] = None
+    backfillResolver: Optional[str] = None
+    backfillLookbackDuration: Optional[float] = None
+    backfillStartTime: Optional[datetime] = None
+    backfillSchedule: Optional[str] = None
+    continuousResolver: Optional[str] = None
+    continuousBufferDuration: Optional[float] = None
+
+
 @dataclasses_json.dataclass_json
 @dataclass
 class UpsertFeatureGQL:
@@ -100,6 +120,7 @@ class UpsertFeatureGQL:
     featureTimeKind: Optional[UpsertFeatureTimeKindGQL] = None
     etlOfflineToOnline: bool = False
     windowBuckets: Optional[List[float]] = None
+    windowMaterialization: Optional[UpsertWindowMaterializationGQL] = None
 
     tags: Optional[List[str]] = None
     maxStaleness: Optional[str] = None
@@ -208,6 +229,7 @@ class UpsertResolverGQL:
     filename: Optional[str] = None
     sourceLine: Optional[int] = None
     dataSources: Optional[List[Any]] = None
+    dataLineage: Optional[Any] = None
 
 
 @dataclasses_json.dataclass_json
@@ -245,13 +267,38 @@ class UpsertCronQueryGQL:
     output: List[str]  # exploded engine-side
     maxSamples: Optional[int]
     recomputeFeatures: RecomputeFeaturesGQL
-    lowerBound: Optional[datetime]
-    upperBound: Optional[datetime]
+    lowerBound: Optional[datetime]  # deprecated: can't use datetime
+    upperBound: Optional[datetime]  # deprecated: can't use datetime
     tags: Optional[List[str]]
     requiredResolverTags: Optional[List[str]]
     storeOnline: Optional[bool] = True  # None = True
     storeOffline: Optional[bool] = True  # None = True
     incrementalSources: Optional[List[str]] = None
+    lowerBoundStr: Optional[str] = None
+    upperBoundStr: Optional[str] = None
+    resourceGroup: Optional[str] = None
+    plannerOptions: Optional[Dict[str, str]] = None
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class UpsertNamedQueryGQL:
+    name: str
+    filename: str
+    queryVersion: Optional[str] = None
+    input: Optional[List[str]] = None
+    output: Optional[List[str]] = None
+    additionalLoggedFeatures: Optional[List[str]] = None
+    tags: Optional[List[str]] = None
+    description: Optional[str] = None
+    owner: Optional[str] = None
+    meta: Optional[Dict[str, str]] = None
+    staleness: Optional[Dict[str, str]] = None
+    plannerOptions: Optional[Dict[str, str]] = None
+    code: Optional[str] = None
+    sourceLineStart: Optional[int] = None
+    sourceLineEnd: Optional[int] = None
+    validPlanNotRequired: Optional[bool] = True
 
 
 @dataclasses_json.dataclass_json
@@ -288,6 +335,7 @@ class EnvironmentSettingsGQL:
     requirements: Optional[str]
     dockerfile: Optional[str]
     requiresPackages: Optional[List[str]] = None
+    platformVersion: Optional[str] = None
 
 
 @dataclasses_json.dataclass_json
@@ -547,7 +595,7 @@ class FeatureClassGQL:
 
 
 @dataclasses_json.dataclass_json
-@dataclass
+@dataclass(frozen=True)
 class PositionGQL:
     """Mirrors an LSP Position
 
@@ -562,7 +610,7 @@ class PositionGQL:
 
 
 @dataclasses_json.dataclass_json
-@dataclass
+@dataclass(frozen=True)
 class RangeGQL:
     """Mirrors an LSP Range
 
@@ -677,6 +725,7 @@ class UpsertGraphGQL:
     streams: Optional[List[UpsertStreamResolverGQL]] = None
     sinks: Optional[List[UpsertSinkResolverGQL]] = None
     cronQueries: Optional[List[UpsertCronQueryGQL]] = None
+    namedQueries: Optional[List[UpsertNamedQueryGQL]] = None
     charts: Optional[List[CreateChartGQL]] = None
     config: Optional[ProjectSettingsGQL] = None
     failed: Optional[List[FailedImport]] = None
