@@ -13,7 +13,6 @@ from chalk._version import __version__
 from chalk.config.project_config import ProjectSettings, load_project_config
 from chalk.features import FeatureSetBase
 from chalk.features.resolver import RESOLVER_REGISTRY
-from chalk.features.scheduled_query import CRON_QUERY_REGISTRY
 from chalk.importer import CHALK_IMPORTER, FailedImport
 from chalk.parsed._graph_validation import validate_graph
 from chalk.parsed.duplicate_input_gql import (
@@ -29,6 +28,7 @@ from chalk.parsed.duplicate_input_gql import (
     ResolverSettings,
     UpsertCronQueryGQL,
     UpsertGraphGQL,
+    UpsertNamedQueryGQL,
     UpsertResolverGQL,
     UpsertSinkResolverGQL,
     UpsertSQLSourceGQL,
@@ -36,6 +36,8 @@ from chalk.parsed.duplicate_input_gql import (
     ValidationSettings,
 )
 from chalk.parsed.json_conversions import convert_type_to_gql, gather_cdc_sources
+from chalk.queries.named_query import NAMED_QUERY_REGISTRY
+from chalk.queries.scheduled_query import CRON_QUERY_REGISTRY
 from chalk.sql._internal.sql_source import BaseSQLSource
 from chalk.utils import paths
 from chalk.utils.paths import get_directory_root
@@ -83,6 +85,7 @@ def project_settings_to_gql(config: ProjectSettings) -> ProjectSettingsGQL:
                             )
                         )
                     ),
+                    platformVersion=e.platform_version,
                 )
                 for i, e in config.environments.items()
             ]
@@ -135,15 +138,18 @@ def get_registered_types(scope_to: Path, failed: List[FailedImport]) -> UpsertGr
             except Exception as e:
                 failed.append(build_failed_import(e, f"feature class '{x.namespace}'"))
                 continue
-            for feature in x.features:
-                if not feature.is_autogenerated and not feature.no_display:
-                    try:
-                        converted = convert_type_to_gql(feature)
-                        features.append(converted)
-                    except Exception as e:
-                        if not LSPErrorBuilder.promote_exception(e):
-                            failed.append(build_failed_import(e, f"feature '{feature.fqn}'"))
-
+            try:
+                for feature in x.features:
+                    if not feature.is_autogenerated and not feature.no_display:
+                        try:
+                            converted = convert_type_to_gql(feature)
+                            features.append(converted)
+                        except Exception as e:
+                            if not LSPErrorBuilder.promote_exception(e):
+                                failed.append(build_failed_import(e, f"feature '{feature.fqn}'"))
+            except Exception as e:
+                if not LSPErrorBuilder.promote_exception(e):
+                    failed.append(build_failed_import(e, f"feature class '{x.namespace}'"))
     project_path = get_directory_root()
     path_prefix_to_remove = project_path.as_posix() if project_path is not None else None
 
@@ -180,6 +186,16 @@ def get_registered_types(scope_to: Path, failed: List[FailedImport]) -> UpsertGr
             except Exception as e:
                 failed.append(build_failed_import(e, f"cron query '{cron_query.name}'"))
 
+    named_queries: list[UpsertNamedQueryGQL] = []
+    for named_query in NAMED_QUERY_REGISTRY.values():
+        named_query.input
+        named_query.output
+        if named_query.filename and _is_relative_to(Path(named_query.filename), scope_to):
+            try:
+                named_queries.append(convert_type_to_gql(named_query, path_prefix=path_prefix_to_remove))
+            except Exception as e:
+                failed.append(build_failed_import(e, f"named query '{named_query.name}'"))
+
     charts: list[CreateChartGQL] = []
     for chart in Chart.registry:
         try:
@@ -199,6 +215,7 @@ def get_registered_types(scope_to: Path, failed: List[FailedImport]) -> UpsertGr
         failed=failed,
         resolvers=resolvers,
         cronQueries=cron_queries,
+        namedQueries=named_queries,
         charts=charts,
         chalkpy=ChalkPYInfo(
             version=__version__,
@@ -219,7 +236,12 @@ def get_registered_types(scope_to: Path, failed: List[FailedImport]) -> UpsertGr
     # Validating the graph produces errors, so do this part last.
     lsp = get_lsp_gql()
     if LSPErrorBuilder.lsp:
-        lsp.diagnostics = CHALK_IMPORTER.supplement_diagnostics(failed_imports=failed, diagnostics=lsp.diagnostics)
+        try:
+            lsp.diagnostics = CHALK_IMPORTER.supplement_diagnostics(failed_imports=failed, diagnostics=lsp.diagnostics)
+        except Exception as e:
+            if len(lsp.diagnostics) == 0 and len(failed) == 0:
+                raise e
+
     graph.lsp = lsp
 
     return graph
