@@ -7,7 +7,9 @@ import json
 import logging
 import os
 import os.path
+import traceback
 import warnings
+from contextvars import ContextVar
 from dataclasses import dataclass
 from enum import Enum
 from os import PathLike
@@ -31,6 +33,7 @@ from typing import (
 )
 
 import pyarrow as pa
+import pyarrow.compute as pc
 
 from chalk.clogging import chalk_logger
 from chalk.features import Feature, FeatureConverter, Features, FeatureWrapper, unwrap_feature
@@ -66,6 +69,10 @@ CHALK_QUERY_LOGGING = env_var_bool("CHALK_QUERY_LOGGING")
 
 _logger = get_logger(__name__)
 
+_ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES: ContextVar[bool] = ContextVar(
+    "_ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES", default=True
+)
+
 
 @dataclass
 class TableIngestionPreferences:
@@ -97,7 +104,7 @@ def validate_dtypes_for_efficient_execution(stmt: Select, supported_types: Seque
         if x["type"].__class__ not in supported_types
     ]
     if len(unsupported_columns) == 1 and unsupported_columns[0]["name"] == "*":
-        _logger.warn(
+        _logger.warning(
             "Got '*' for select clause. Will try efficient execution and fallback if unsupported types are found."
         )
         return
@@ -149,15 +156,27 @@ class TableIngestMixIn(TableIngestProtocol):
 
 class SQLSourceKind(str, Enum):
     bigquery = "bigquery"
+    clickhouse = "clickhouse"
     cloudsql = "cloudsql"
     databricks = "databricks"
+    athena = "athena"
+    duckdb = "duckdb"
+    dynamodb = "dynamodb"
     mysql = "mysql"
     postgres = "postgres"
     redshift = "redshift"
     snowflake = "snowflake"
+    spanner = "spanner"
     sqlite = "sqlite"
     trino = "trino"
 
+    @staticmethod
+    def convert_sql_source_kind(kind: str) -> SQLSourceKind:
+        try:
+            return SQLSourceKind(kind)
+        except ValueError:
+            raise ValueError(f"Unsupported SQL source kind: {kind}")
+
 
 class BaseSQLSource(BaseSQLSourceProtocol):
     registry: ClassVar[List["BaseSQLSource"]] = []
@@ -179,31 +198,53 @@ class BaseSQLSource(BaseSQLSourceProtocol):
 
         self._incremental_settings = None
         self._resolver_and_sqlfile_to_sqlstring: Dict[Tuple[str, str], str] = {}
-        self.registry.append(self)
+        if _ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES.get():
+            self.registry.append(self)
         self.name = name
         if engine_args is None:
             engine_args = {}
         if async_engine_args is None:
             async_engine_args = {}
         if self.name is not None:
-            for k, v in self._load_env_engine_args(name=self.name).items():
+            for k, v in self._load_env_engine_args(name=self.name, override=None).items():
                 engine_args.setdefault(k, v)
                 async_engine_args.setdefault(k, v)
         if getattr(self, "kind", None) != SQLSourceKind.trino:
             engine_args.setdefault("pool_pre_ping", env_var_bool("USE_CLIENT_POOL_PRE_PING"))
-            engine_args.setdefault("echo", CHALK_QUERY_LOGGING)
             async_engine_args.setdefault("pool_pre_ping", env_var_bool("USE_CLIENT_POOL_PRE_PING"))
-            async_engine_args.setdefault("echo", CHALK_QUERY_LOGGING)
-        self._engine_args = engine_args
-        self._async_engine_args = async_engine_args
+        self.engine_args = engine_args
+        self.async_engine_args = async_engine_args
         self._engine = None
         self._async_engine = None
 
+    @property
+    def _engine_args(self):
+        """Backcompat support for private subclassing of BaseSQLSource"""
+        return self.engine_args
+
+    @_engine_args.setter
+    def _engine_args(self, args: dict[str, Any]):
+        """Backcompat support for private subclassing of BaseSQLSource"""
+        self.engine_args = args
+
+    @property
+    def _async_engine_args(self):
+        """Backcompat support for private subclassing of BaseSQLSource"""
+        return self.async_engine_args
+
+    @_async_engine_args.setter
+    def _async_engine_args(self, args: dict[str, Any]):
+        """Backcompat support for private subclassing of BaseSQLSource"""
+        self.async_engine_args = args
+
     def get_sqlglot_dialect(self) -> Union[str, None]:
         """Returns the name of the SQL dialect (if it has one) for `sqlglot` to parse the SQL string.
         This allows for use of dialect-specific syntax while parsing and modifying queries."""
         return None
 
+    def supports_inefficient_fallback(self) -> bool:
+        return True
+
     def query_sql_file(
         self,
         path: Union[str, bytes, PathLike],
@@ -228,6 +269,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                 frame = inspect.currentframe()
                 assert frame is not None
                 caller_frame = frame.f_back
+                del frame
                 assert caller_frame is not None
                 while caller_frame.f_globals["__name__"].startswith("chalk.sql") and caller_frame.f_back is not None:
                     caller_frame = caller_frame.f_back
@@ -303,7 +345,21 @@ class BaseSQLSource(BaseSQLSourceProtocol):
             passed in. Thus, we need to hack the env vars in.
             """
             query_execution_parameters = query_execution_parameters_from_env_vars()
+
+        efficient_execution_failed: bool = False
+        efficient_execution_traceback: str | None = None
         attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution
+
+        if CHALK_QUERY_LOGGING:
+            log_dialect = None
+            try:
+                log_dialect = self.get_sqlalchemy_dialect(paramstyle="pyformat")
+            except Exception as e:
+                _logger.warning(f"Could not determine sqlalchemy dialect of source of type {self.kind}", exc_info=e)
+            _logger.info(
+                f"Executing SQL query: {finalized_query.query.compile(compile_kwargs={'literal_binds': True}, dialect=log_dialect)}"
+            )
+
         if attempt_efficient_execution:
             try:
                 yield from self._execute_query_efficient(
@@ -322,7 +378,10 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                 log_level = e.log_level
                 _logger.log(log_level, str(e))
             except Exception:
-                if query_execution_parameters.fallback_to_inefficient_execution:
+                if (
+                    query_execution_parameters.fallback_to_inefficient_execution
+                    and self.supports_inefficient_fallback()
+                ):
                     _logger.error(
                         (
                             f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
@@ -330,14 +389,25 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                         ),
                         exc_info=True,
                     )
+                    efficient_execution_failed = True
+                    efficient_execution_traceback = traceback.format_exc()
                 else:
                     raise
-
-        batch = self._execute_query_inefficient(
-            finalized_query, columns_to_features, connection, query_execution_parameters
-        )
-        if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
-            yield batch
+        try:
+            batch = self._execute_query_inefficient(
+                finalized_query, columns_to_features, connection, query_execution_parameters
+            )
+        except:
+            raise
+        else:
+            if efficient_execution_failed:
+                # This helps us collect into about the causes of errors in the 'efficient' code path.
+                # i.e. are they code bugs we can fix, or actual errors e.g. in the database itself?
+                _logger.info(
+                    f"Failed to efficiently execute query, but inefficient code path succeeded. Original error was: {efficient_execution_traceback}"
+                )
+            if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
+                yield batch
 
     @contextlib.contextmanager
     def _create_temp_table(
@@ -399,6 +469,8 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         if query_execution_parameters is None:
             query_execution_parameters = query_execution_parameters_from_env_vars()
         attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution
+        efficient_execution_failed: bool = False
+        efficient_execution_traceback: str | None = None
         if attempt_efficient_execution:
             try:
                 async for res in self._async_execute_query_efficient(
@@ -438,6 +510,8 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                             ),
                             exc_info=True,
                         )
+                        efficient_execution_failed = True
+                        efficient_execution_traceback = traceback.format_exc()
                     else:
                         raise
             except UnsupportedEfficientExecutionError as e:
@@ -452,14 +526,25 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                         ),
                         exc_info=True,
                     )
+                    efficient_execution_failed = True
+                    efficient_execution_traceback = traceback.format_exc()
                 else:
                     raise
-
-        batch = await self._async_execute_query_inefficient(
-            finalized_query, columns_to_features, connection, query_execution_parameters
-        )
-        if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
-            yield batch
+        try:
+            batch = await self._async_execute_query_inefficient(
+                finalized_query, columns_to_features, connection, query_execution_parameters
+            )
+        except:
+            raise
+        else:
+            if efficient_execution_failed:
+                # This helps us collect into about the causes of errors in the 'efficient' code path.
+                # i.e. are they code bugs we can fix, or actual errors e.g. in the database itself?
+                _logger.info(
+                    f"Failed to efficiently execute query, but inefficient code path succeeded. Original error was: {efficient_execution_traceback}"
+                )
+            if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
+                yield batch
 
     async def _async_execute_query_inefficient(
         self,
@@ -524,8 +609,9 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         query_execution_parameters: QueryExecutionParameters,
     ) -> pa.RecordBatch:
         with (
-            self.get_engine().connect() if connection is None else contextlib.nullcontext(connection)
-        ) as cnx, cnx.begin():
+            self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx,
+            cnx.begin(),
+        ):
             with contextlib.ExitStack() as exit_stack:
                 for (
                     _,
@@ -599,12 +685,12 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         compiled_query = query.compile(dialect=dialect)
         return compiled_query
 
-    def _load_env_engine_args(self, name: str) -> Mapping[str, Any]:
+    def _load_env_engine_args(self, name: str, override: Optional[Mapping[str, str]]) -> Mapping[str, Any]:
         """
         Loads additional engine arguments from env var "{name}_ENGINE_ARGUMENTS"
         """
 
-        extra_args = load_integration_variable(integration_name=name, name="ENGINE_ARGUMENTS")
+        extra_args = load_integration_variable(integration_name=name, name="ENGINE_ARGUMENTS", override=override)
         if extra_args is None:
             return {}
         else:
@@ -613,7 +699,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
             return extra_args
 
     def _check_engine_isolation_level(self):
-        isolation_level = self._engine_args.get("isolation_level")
+        isolation_level = self.engine_args.get("isolation_level")
         if isolation_level == "AUTOCOMMIT":
             warnings.warn(
                 UserWarning(
@@ -634,7 +720,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         if self._engine is None:
             self.register_sqlalchemy_compiler_overrides()
             self._check_engine_isolation_level()
-            self._engine = create_engine(url=self.local_engine_url(), **self._engine_args)
+            self._engine = create_engine(url=self.local_engine_url(), **self.engine_args)
         return self._engine
 
     def get_async_engine(self):
@@ -643,7 +729,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         if self._async_engine is None:
             self.register_sqlalchemy_compiler_overrides()
             self._check_engine_isolation_level()
-            self._async_engine = create_async_engine(url=self.async_local_engine_url(), **self._async_engine_args)
+            self._async_engine = create_async_engine(url=self.async_local_engine_url(), **self.async_engine_args)
         return self._async_engine
 
     def raw_session(self) -> Session:
@@ -674,6 +760,9 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         """Hook to register SQLAlchemy Compiler overrides. These hooks are registered by default in tests and in the engine"""
         return
 
+    def _recreate_integration_variables(self) -> dict[str, str]:
+        raise NotImplementedError
+
 
 def _extract_features(feature_set: Features, prefix: str, targets: List[Label], features: Dict[str, Feature]):
     import sqlalchemy.sql.functions
@@ -717,6 +806,76 @@ def _extract_features(feature_set: Features, prefix: str, targets: List[Label],
 insert_size_limit = 10_000
 
 
+def _replace_float16_with_float_64(t: pa.DataType) -> pa.DataType:
+    if t == pa.float16():
+        return pa.float64()
+    if pa.types.is_fixed_size_list(t):
+        elem_type = t.field(0).type
+        elem_type_replaced = _replace_float16_with_float_64(elem_type)
+        if elem_type_replaced is elem_type:
+            return t
+        return pa.list_(elem_type_replaced, list_size=t.list_size)
+    if pa.types.is_large_list(t):
+        elem_type = t.field(0).type
+        elem_type_replaced = _replace_float16_with_float_64(elem_type)
+        if elem_type_replaced is elem_type:
+            return t
+        return pa.large_list(elem_type_replaced)
+    if pa.types.is_list(t):
+        elem_type = t.field(0).type
+        elem_type_replaced = _replace_float16_with_float_64(elem_type)
+        if elem_type_replaced is elem_type:
+            return t
+        return pa.list_(elem_type_replaced)
+    if pa.types.is_struct(t):
+        new_elems: List[pa.Field] = []
+        any_new_elem = False
+        for field_index in range(t.num_fields):
+            f = t.field(field_index)
+            elem_replaced = _replace_float16_with_float_64(f.type)
+            if elem_replaced is f.type:
+                new_elems.append(f)
+            else:
+                any_new_elem = True
+                new_elems.append(f.with_type(elem_replaced))
+
+        if not any_new_elem:
+            return t
+
+        return pa.struct(new_elems)
+
+    if pa.types.is_map(t):
+        as_struct = t.field(0).type
+        new_value_type = _replace_float16_with_float_64(as_struct.field(1).type)
+        if new_value_type is as_struct.field(1).type:
+            return t
+        return pa.map_(as_struct.field(0), as_struct.field(1).with_type(new_value_type))
+
+    return t
+
+
+def _convert_to_arrow_array(
+    data: List[Any],
+    *,
+    type: pa.DataType,
+    field_name: str,
+) -> pa.Array:
+    try:
+        return pa.array(data, type=type)
+    except Exception as e:
+        # pyarrow<=20.0 has a bug where it cannot read in float16 values from regular `float`s.
+
+        type_without_float16 = _replace_float16_with_float_64(type)
+        if type_without_float16 is not type:
+            try:
+                regular_float_array = pa.array(data, type=type_without_float16)
+                return pc.cast(regular_float_array, type)
+            except Exception as e:
+                raise ValueError(f"Unable to convert values for feature '{field_name}' into pyarrow type '{type}': {e}")
+
+        raise ValueError(f"Unable to convert values for feature '{field_name}' into pyarrow type '{type}': {e}")
+
+
 def _convert_to_record_batch(
     data: Mapping[str, List[Any]],
     schema: pa.Schema,
@@ -728,10 +887,9 @@ def _convert_to_record_batch(
     """
     data_arrays: dict[str, pa.Array] = dict()
     for k, v in data.items():
-        try:
-            data_arrays[k] = pa.array(v, type=schema.field(k).type)
-        except Exception as e:
-            raise ValueError(
-                f"Unable to convert values for feature '{k}' into pyarrow type '{schema.field(k).type}': {e}"
-            )
+        data_arrays[k] = _convert_to_arrow_array(
+            v,
+            type=schema.field(k).type,
+            field_name=k,
+        )
     return pa.RecordBatch.from_pydict(data_arrays, schema=schema)
