@@ -1,21 +1,30 @@
-from typing import TYPE_CHECKING, Any, List, Literal, Optional, Union
+from typing import TYPE_CHECKING, Any, List, Literal, Mapping, Optional, Union
 
-from chalk.integrations.named import load_integration_variable
+from chalk.integrations.named import create_integration_variable, load_integration_variable
 from chalk.sink._models import SinkIntegrationProtocol
 from chalk.streams.base import StreamSource
 from chalk.utils.duration import Duration
-from chalk.utils.string import comma_whitespace_split
+from chalk.utils.string import comma_join, comma_whitespace_split
 
 if TYPE_CHECKING:
-    import pydantic
-    from pydantic import BaseModel
+    from pydantic import BaseModel, Field
 else:
     try:
-        import pydantic.v1 as pydantic
-        from pydantic.v1 import BaseModel
+        from pydantic.v1 import BaseModel, Field
     except ImportError:
-        import pydantic
-        from pydantic import BaseModel
+        from pydantic import BaseModel, Field
+
+
+_KAFKA_BOOTSTRAP_SERVER_NAME = "KAFKA_BOOTSTRAP_SERVER"
+_KAFKA_TOPIC_NAME = "KAFKA_TOPIC"
+_KAFKA_SSL_KEYSTORE_LOCATION_NAME = "KAFKA_SSL_KEYSTORE_LOCATION"
+_KAFKA_SSL_CA_FILE_NAME = "KAFKA_SSL_CA_FILE"
+_KAFKA_CLIENT_ID_PREFIX_NAME = "KAFKA_CLIENT_ID_PREFIX"
+_KAFKA_GROUP_ID_PREFIX_NAME = "KAFKA_GROUP_ID_PREFIX"
+_KAFKA_SECURITY_PROTOCOL_NAME = "KAFKA_SECURITY_PROTOCOL"
+_KAFKA_SASL_MECHANISM_NAME = "KAFKA_SASL_MECHANISM"
+_KAFKA_SASL_USERNAME_NAME = "KAFKA_SASL_USERNAME"
+_KAFKA_SASL_PASSWORD_NAME = "KAFKA_SASL_PASSWORD"
 
 
 class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True):
@@ -63,7 +72,7 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
     Username for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
     """
 
-    sasl_password: Optional[str] = pydantic.Field(default=None, repr=False)
+    sasl_password: Optional[str] = Field(default=None, repr=False)
     """
     Password for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
     """
@@ -99,33 +108,59 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
         name: Optional[str] = None,
         late_arrival_deadline: Duration = "infinity",
         dead_letter_queue_topic: Optional[str] = None,
+        integration_variable_override: Optional[Mapping[str, str]] = None,
     ):
         super(KafkaSource, self).__init__(
             bootstrap_server=bootstrap_server
             or load_integration_variable(
-                name="KAFKA_BOOTSTRAP_SERVER", integration_name=name, parser=comma_whitespace_split
+                name=_KAFKA_BOOTSTRAP_SERVER_NAME,
+                integration_name=name,
+                parser=comma_whitespace_split,
+                override=integration_variable_override,
+            ),
+            topic=topic
+            or load_integration_variable(
+                name=_KAFKA_TOPIC_NAME, integration_name=name, override=integration_variable_override
             ),
-            topic=topic or load_integration_variable(name="KAFKA_TOPIC", integration_name=name),
             ssl_keystore_location=ssl_keystore_location
-            or load_integration_variable(name="KAFKA_SSL_KEYSTORE_LOCATION", integration_name=name),
+            or load_integration_variable(
+                name=_KAFKA_SSL_KEYSTORE_LOCATION_NAME, integration_name=name, override=integration_variable_override
+            ),
             client_id_prefix=client_id_prefix
-            or load_integration_variable(name="KAFKA_CLIENT_ID_PREFIX", integration_name=name)
+            or load_integration_variable(
+                name=_KAFKA_CLIENT_ID_PREFIX_NAME, integration_name=name, override=integration_variable_override
+            )
             or KafkaSource.__fields__["client_id_prefix"].default,
             group_id_prefix=group_id_prefix
-            or load_integration_variable(name="KAFKA_GROUP_ID_PREFIX", integration_name=name)
+            or load_integration_variable(
+                name=_KAFKA_GROUP_ID_PREFIX_NAME, integration_name=name, override=integration_variable_override
+            )
             or KafkaSource.__fields__["group_id_prefix"].default,
             security_protocol=security_protocol
-            or load_integration_variable(name="KAFKA_SECURITY_PROTOCOL", integration_name=name)
+            or load_integration_variable(
+                name=_KAFKA_SECURITY_PROTOCOL_NAME, integration_name=name, override=integration_variable_override
+            )
             or KafkaSource.__fields__["security_protocol"].default,
             sasl_mechanism=sasl_mechanism
-            or load_integration_variable(name="KAFKA_SASL_MECHANISM", integration_name=name)
+            or load_integration_variable(
+                name=_KAFKA_SASL_MECHANISM_NAME, integration_name=name, override=integration_variable_override
+            )
             or KafkaSource.__fields__["sasl_mechanism"].default,
-            sasl_username=sasl_username or load_integration_variable(name="KAFKA_SASL_USERNAME", integration_name=name),
-            sasl_password=sasl_password or load_integration_variable(name="KAFKA_SASL_PASSWORD", integration_name=name),
+            sasl_username=sasl_username
+            or load_integration_variable(
+                name=_KAFKA_SASL_USERNAME_NAME, integration_name=name, override=integration_variable_override
+            ),
+            sasl_password=sasl_password
+            or load_integration_variable(
+                name=_KAFKA_SASL_PASSWORD_NAME, integration_name=name, override=integration_variable_override
+            ),
             name=name,
             late_arrival_deadline=late_arrival_deadline,
             dead_letter_queue_topic=dead_letter_queue_topic,
-            ssl_ca_file=ssl_ca_file or load_integration_variable(name="KAFKA_SSL_CA_FILE", integration_name=name),
+            ssl_ca_file=ssl_ca_file
+            or load_integration_variable(
+                name=_KAFKA_SSL_CA_FILE_NAME, integration_name=name, override=integration_variable_override
+            ),
         )
         self.registry.append(self)
 
@@ -144,3 +179,23 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
     def stream_or_topic_name(self):
         assert self.topic is not None
         return self.topic
+
+    def _recreate_integration_variables(self) -> dict[str, str]:
+        return {
+            k: v
+            for k, v in [
+                create_integration_variable(
+                    _KAFKA_BOOTSTRAP_SERVER_NAME, self.name, self.bootstrap_server, serializer=comma_join
+                ),
+                create_integration_variable(_KAFKA_TOPIC_NAME, self.name, self.topic),
+                create_integration_variable(_KAFKA_SSL_KEYSTORE_LOCATION_NAME, self.name, self.ssl_keystore_location),
+                create_integration_variable(_KAFKA_SSL_CA_FILE_NAME, self.name, self.ssl_ca_file),
+                create_integration_variable(_KAFKA_CLIENT_ID_PREFIX_NAME, self.name, self.client_id_prefix),
+                create_integration_variable(_KAFKA_GROUP_ID_PREFIX_NAME, self.name, self.group_id_prefix),
+                create_integration_variable(_KAFKA_SECURITY_PROTOCOL_NAME, self.name, self.security_protocol),
+                create_integration_variable(_KAFKA_SASL_MECHANISM_NAME, self.name, self.sasl_mechanism),
+                create_integration_variable(_KAFKA_SASL_USERNAME_NAME, self.name, self.sasl_username),
+                create_integration_variable(_KAFKA_SASL_PASSWORD_NAME, self.name, self.sasl_password),
+            ]
+            if v is not None
+        }
