@@ -1,7 +1,9 @@
 from __future__ import annotations
 
 import io
+import math
 import pathlib
+import typing
 from io import BytesIO
 from typing import TYPE_CHECKING, Any, BinaryIO, Literal, Union
 
@@ -9,6 +11,8 @@ import numpy as np
 import pyarrow as pa
 import pyarrow.compute as pc
 import pyarrow.ipc as pa_ipc
+import pyarrow.parquet as pq
+import requests
 
 from chalk.features._encoding.pyarrow import is_map_in_dtype_tree, pyarrow_to_polars
 from chalk.utils.log_with_context import get_logger
@@ -20,6 +24,8 @@ if TYPE_CHECKING:
 
 _logger = get_logger(__name__)
 
+T_ArrowTable = typing.TypeVar("T_ArrowTable", bound=typing.Union[pa.RecordBatch, pa.Table])
+
 
 class ArrowTableCastError(Exception):
     def __init__(self, errors: list[ArrowColumnCastError]):
@@ -53,13 +59,23 @@ class PyArrowToPolarsConverter:
     can be cached
     """
 
-    def __init__(self, schema: pa.Schema) -> None:
+    def __init__(self, schema: pa.Schema, use_fixed_size_list: bool = False) -> None:
         super().__init__()
         self.schema = schema
+        self.no_ext_schema = self._remove_extension_types(schema)
         self._pl_schema = {
-            col_name: pyarrow_to_polars(schema.field(col_name).type, col_name) for col_name in schema.names
+            col_name: pyarrow_to_polars(schema.field(col_name).type, col_name, use_fixed_size_list=use_fixed_size_list)
+            for col_name in schema.names
         }
 
+    def _does_dtype_have_float16(self, dtype: pa.DataType) -> bool:
+        if pa.types.is_float16(dtype):
+            return True
+        for fidx in range(dtype.num_fields):
+            if self._does_dtype_have_float16(dtype.field(fidx).type):
+                return True
+        return False
+
     def _recursive_convert_map(self, arr: Union[pa.Array, pa.ChunkedArray]) -> pa.Array:
         """
         Recursively converts a `MapArray` to a `LargeList` of `StructArray`s,
@@ -137,40 +153,70 @@ class PyArrowToPolarsConverter:
         else:
             return arr
 
+    @classmethod
+    def _recursive_convert_float16(cls, x: pa.Array) -> pa.Array:
+        """
+        Recursively replace float16 columns with float32 since Polars doesn't support float16.
+        1. If col is float16, cast to float32
+        2. If col is a fixed-size list of float16's, convert the underlying values to float32 and fill in nulls w/ lists of null values of the apprp length.
+        :param x:
+        :return:
+        """
+        if pa.types.is_float16(x.type):
+            return pa.array(x.to_numpy(zero_copy_only=False).astype(np.dtype("float32")))
+        if x.type.num_fields == 0:
+            # if it's not a nested type and it's not float16, return as is
+            return x
+
+        if pa.types.is_fixed_size_list(x.type):
+            assert isinstance(x.type, pa.FixedSizeListType)
+            if pa.types.is_float16(x.type.value_type):
+                assert isinstance(x, pa.FixedSizeListArray)
+                # We'll first expand all null elements into null lists of the correct length, and then convert to numpy
+                null_elements = x.is_null()
+                empty = pa.scalar(
+                    np.zeros((x.type.list_size,), dtype=np.dtype("float16")),
+                    pa.list_(pa.float16(), x.type.list_size),
+                )
+                x = x.fill_null(empty)
+                pa_arr = pa.FixedSizeListArray.from_arrays(
+                    x.flatten().to_numpy(zero_copy_only=False).astype(np.dtype("float32")),
+                    x.type.list_size,
+                )
+                # Replace the filled empty elements with null
+                pa_arr = pc.if_else(  # type: ignore
+                    null_elements, pa.scalar(None, pa.list_(pa.float32(), x.type.list_size)), pa_arr
+                )
+                return pa_arr
+            # TODO: nested fixed-size-lists e.g. FixedSizeList<FixedSizeList<float16>>
+        if pa.types.is_struct(x.type):
+            assert isinstance(x, pa.StructArray)
+            child_arrays = x.flatten()
+            child_arrays = [cls._recursive_convert_float16(x) for x in child_arrays]
+            child_names = [x.name for x in x.type.fields]
+            converted_struct_array = pa.StructArray.from_arrays(
+                arrays=child_arrays, names=child_names, mask=x.is_null()
+            )
+            # Apply the original null mask to new struct array
+            return pc.if_else(pc.is_null(x), None, converted_struct_array)  # type: ignore
+        if pa.types.is_list(x.type) or pa.types.is_large_list(x.type):
+            values_converted = x.flatten()
+            values_converted = cls._recursive_convert_float16(values_converted)
+            # (call class method w/ the object to get the correct sub-class)
+            return x.from_arrays(offsets=x.offsets, values=values_converted, mask=x.is_null())
+        return x
+
     def convert(self, table: pa.Table) -> pl.DataFrame:
         import polars as pl
 
         assert table.schema == self.schema, "The table schema differs from the declared schema"
 
-        # Polars cannot handle pa.float16, nor can pyarrow cast float16 to float32. So, we will manually cast float16 to float32 inside numpy
-        # FIXME: While this should be recursive, in practice, we only encounter float16 inside of vector features
+        table = table.cast(self.no_ext_schema)
+
         new_cols: list[pa.Array | pa.ChunkedArray] = []
         for x in table.columns:
-            if pa.types.is_float16(x.type):
-                new_cols.append(pa.array(x.to_numpy(zero_copy_only=False).astype(np.dtype("float32"))))
-                continue
-            if pa.types.is_fixed_size_list(x.type):
-                assert isinstance(x.type, pa.FixedSizeListType)
-                if pa.types.is_float16(x.type.value_type):
-                    as_array = x.combine_chunks()
-                    assert isinstance(as_array, pa.FixedSizeListArray)
-                    # We'll first expand all null elements into null lists of the correct length, and then convert to numpy
-                    null_elements = as_array.is_null()
-                    empty = pa.scalar(
-                        np.empty((x.type.list_size,), dtype=np.dtype("float16")),
-                        pa.list_(pa.float16(), x.type.list_size),
-                    )
-                    as_array = as_array.fill_null(empty)
-                    pa_arr = pa.FixedSizeListArray.from_arrays(
-                        as_array.flatten().to_numpy(zero_copy_only=False).astype(np.dtype("float32")),
-                        x.type.list_size,
-                    )
-                    # Replace the filled empty elements with null
-                    pa_arr = pc.if_else(  # type: ignore
-                        null_elements, pa.scalar(None, pa.list_(pa.float32(), x.type.list_size)), pa_arr
-                    )
-                    new_cols.append(pa_arr)
-                    continue
+            if self._does_dtype_have_float16(x.type):
+                x = self._recursive_convert_float16(x.combine_chunks())
             if is_map_in_dtype_tree(x.type):
                 new_cols.append(self._recursive_convert_map(x))
                 continue
@@ -209,6 +255,25 @@ class PyArrowToPolarsConverter:
             df = df.with_columns(list(col_name_to_expr.values()))
         return df
 
+    @staticmethod
+    def _source_type_of_extension_type(dtype: pa.DataType) -> pa.DataType:
+        if isinstance(dtype, pa.ExtensionType):
+            return dtype.storage_type
+        return dtype
+
+    @staticmethod
+    def _remove_extension_types(table: pa.Schema) -> pa.Schema:
+        return pa.schema(
+            [
+                pa.field(
+                    name=field.name,
+                    type=PyArrowToPolarsConverter._source_type_of_extension_type(field.type),
+                    metadata=field.metadata,
+                )
+                for field in table
+            ]
+        )
+
 
 def pa_table_to_pl_df(table: pa.Table[Any]) -> pl.DataFrame:
     return PyArrowToPolarsConverter(table.schema).convert(table)
@@ -254,19 +319,18 @@ def pa_cast(table: pa.Table[Any], expected_schema: pa.Schema, collect_all_errors
             if len(chunk) == 0:
                 continue
             try:
-                chunk_res = _pa_cast_col(chunk, expected_type)
+                chunk_res = pa_cast_col(chunk, expected_type)
             except Exception as e:
-                if not collect_all_errors:
-                    raise e
-                errors.append(
-                    ArrowColumnCastError(
-                        msg=str(e), col_name=name, expected_dtype=expected_type, actual_dtype=chunk.type
-                    )
+                wrapped_error = ArrowColumnCastError(
+                    msg=str(e), col_name=name, expected_dtype=expected_type, actual_dtype=chunk.type
                 )
+                if not collect_all_errors:
+                    raise wrapped_error from e
+                errors.append(wrapped_error)
             else:
                 casted_chunks.append(chunk_res)
         if len(casted_chunks) == 0:
-            arr = pa.array([], expected_type)
+            arr = pa.nulls(0, expected_type)
         else:
             arr = pa.chunked_array(casted_chunks)
         new_arrays.append(arr)
@@ -289,7 +353,7 @@ def recursive_convert_map_primitive(x: Any, dtype: pa.DataType):
     if isinstance(dtype, pa.MapType):
         if not isinstance(x, list):
             raise ValueError(f"Expected a list, but got {type(x).__name__}")
-        return {e["key"]: recursive_convert_map_primitive(e["value"], dtype.item_type) for e in x}
+        return {e[0]: recursive_convert_map_primitive(e[1], dtype.item_type) for e in x}
     elif isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType)):
         if not isinstance(x, list):
             raise ValueError(f"Expected a list, but got {type(x).__name__}")
@@ -343,15 +407,34 @@ def _recursive_convert_map_type(dtype: pa.DataType) -> pa.DataType:
     return dtype
 
 
-def _pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
+def pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
+    """
+    Converts the provided input type to the provided expected type.
+    This conversion is used to avoid hard-to-represent types, like fixed-sized-lists or maps, which
+    polars in particular does not support.
+    """
     if col.null_count == len(col):
         # It's all null, so short circuit and return an array of nulls of the correct type
         # calling .flatten() sometimes will raise an exception if this condition is true
         return pa.nulls(len(col), expected_type)
-    if is_map_in_dtype_tree(expected_type):
-        expected_type = _recursive_convert_map_type(expected_type)
     if col.type == expected_type:
         return col
+    if pa.types.is_map(expected_type):
+        assert isinstance(expected_type, pa.MapType)
+        # Validate that the arrow "maps" are in list<struct<key,value>> format
+        if not (
+            (isinstance(col, pa.LargeListArray) or isinstance(col, pa.ListArray))
+            and hasattr(col, "values")
+            and isinstance(col.values, pa.StructArray)
+            and isinstance(struct := col.values.type, pa.StructType)
+            and struct.get_field_index("key") == 0
+            and struct.get_field_index("value") == 1
+        ):
+            raise ValueError(f"Expected the column to be a list of <key,value> structs, instead got {col.type}")
+
+        keys = pa_cast_col(col.values.field("key"), expected_type.key_type)
+        values = pa_cast_col(col.values.field("value"), expected_type.item_type)
+        return pa.MapArray.from_arrays(col.offsets, keys, values, mask=col.is_null())
     if pa.types.is_struct(expected_type):
         # Convert the column to a table, then recursively cast the table
         assert isinstance(expected_type, pa.StructType)
@@ -375,7 +458,7 @@ def _pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
         return struct_array
     if pa.types.is_list(expected_type) or pa.types.is_large_list(expected_type):
         assert isinstance(expected_type, (pa.LargeListType, pa.ListType))
-        assert isinstance(col, (pa.ListArray, pa.LargeListArray))
+        assert isinstance(col, (pa.ListArray, pa.LargeListArray, pa.FixedSizeListArray))
         flattened = col.flatten()
         tbl = pa.Table.from_arrays([flattened], names=[expected_type.value_field.name])
         tbl = pa_cast(tbl, pa.schema([expected_type.value_field]))
@@ -408,7 +491,6 @@ def _pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
             assert isinstance(offsets, pa.Int32Array)
 
             ans = pa.ListArray.from_arrays(offsets, casted_col)
-
         else:
             assert pa.types.is_large_list(expected_type)
             if not isinstance(offsets, pa.Int64Array):
@@ -419,42 +501,66 @@ def _pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
         assert len(ans) == len(col), "array should have the same number of elements"
         return ans
     if pa.types.is_fixed_size_list(expected_type):
-        assert isinstance(expected_type, pa.FixedSizeListType)
-        # We'll first expand all null elements into null lists of the correct length, and then convert to numpy
-        null_elements = col.is_null()
-        empty = pa.scalar(
-            np.empty((expected_type.list_size,), dtype=expected_type.value_type.to_pandas_dtype()), expected_type
-        )
-        col = col.fill_null(empty)
-        if isinstance(col, (pa.ListArray, pa.LargeListArray)):
-            # Possible if we are coming from polars
-            if pa.types.is_float16(expected_type.value_type):
-                # For float16, we need to go to numpy
-                flattened = pa.array(
-                    col.flatten().to_numpy(zero_copy_only=False).reshape(-1).astype(np.dtype("float16"))
+        # Cast the elements of the input column, and then reconstruct an outer list.
+        assert isinstance(col, (pa.ListArray, pa.LargeListArray, pa.FixedSizeListArray))
+        if isinstance(col, pa.ListArray):
+            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
+            rebuilt_col = pa.ListArray.from_arrays(
+                offsets=col.offsets,
+                values=col_items_converted,
+                type=pa.list_(expected_type.value_type),
+            )
+            if col.null_count != 0:
+                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
+                # To support this case, null out the missing lists with `if_else`.
+
+                # Note: we are missing type stubs for `pc.if_else`
+                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
+                    col.is_null(),
+                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
+                    rebuilt_col,
+                )
+        elif isinstance(col, pa.LargeListArray):
+            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
+            rebuilt_col = pa.LargeListArray.from_arrays(
+                offsets=col.offsets,
+                values=col_items_converted,
+                type=pa.large_list(expected_type.value_type),
+            )
+            if col.null_count != 0:
+                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
+                # To support this case, null out the missing lists with `if_else`.
+
+                # Note: we are missing type stubs for `pc.if_else`
+                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
+                    col.is_null(),
+                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
+                    rebuilt_col,
+                )
+        elif isinstance(col, pa.FixedSizeListArray):
+            col_items_converted = pa_cast_col(col.values, expected_type=expected_type.value_type)
+            rebuilt_col = pa.FixedSizeListArray.from_arrays(
+                values=col_items_converted,
+                type=pa.list_(expected_type.value_type, col.type.list_size),
+            )
+            if col.null_count != 0:
+                # Pyarrow does not allow `mask` to be set if `offset` is non-zero.
+                # To support this case, null out the missing lists with `if_else`.
+
+                # Note: we are missing type stubs for `pc.if_else`
+                rebuilt_col = pc.if_else(  # pyright: ignore[reportAttributeAccessIssue]
+                    col.is_null(),
+                    pa.nulls(len(rebuilt_col), rebuilt_col.type),
+                    rebuilt_col,
                 )
-            else:
-                # For everything else, we go through python
-                # Cannot go through numpy because ints will be cast to floats, since numpy doesn't have null
-                # FIXME: Do this in c and skip python
-                flattened = pa.array(col.to_pylist(), expected_type)
-                assert isinstance(flattened, pa.FixedSizeListArray)
-                flattened = flattened.flatten()
         else:
-            assert isinstance(col.type, pa.FixedSizeListType)
-            assert isinstance(col, pa.FixedSizeListArray)
-            assert col.type.list_size == expected_type.list_size
-            flattened = col.flatten()
-        tbl = pa.Table.from_arrays([flattened], names=[expected_type.value_field.name])
-        tbl = pa_cast(tbl, pa.schema([expected_type.value_field]))
-        casted_col = tbl.column(0).combine_chunks()
-        ans = pa.FixedSizeListArray.from_arrays(casted_col, expected_type.list_size)
-        assert len(ans) == len(col), "array should have the same number of elements"
-        # Replace the filled empty elements with null
-        ans = pc.if_else(null_elements, pa.scalar(None, expected_type), ans)  # type: ignore
-        return ans
+            raise ValueError(
+                f"Cannot convert array with non-list type {col.type} to fixed-sized list type '{expected_type}'"
+            )
+
+        return pc.cast(rebuilt_col, expected_type)
     if pa.types.is_map(expected_type):
-        return _pa_cast_col(col, expected_type)
+        return pa_cast_col(col, expected_type)
 
     # Otherwise, cast directly if it's a scalar
 
@@ -470,6 +576,13 @@ def _pa_cast_col(col: pa.Array, expected_type: pa.DataType) -> pa.Array:
         raise e
 
 
+def pa_replace_column(t: T_ArrowTable, col_name: str, col: pa.Array | pa.ChunkedArray) -> T_ArrowTable:
+    idx = t.schema.get_field_index(col_name)
+    if idx < 0:
+        raise KeyError(f"Could not find unique column with name {col_name} in table.")
+    return t.set_column(idx, col_name, col)
+
+
 def _read_parquet(
     source: str | pathlib.Path | BinaryIO | BytesIO | bytes,
     **kwargs: Any,
@@ -486,7 +599,28 @@ def _read_parquet(
         if source.startswith("gs://"):
             # Fsspec expects gcs instead of gs
             source = "gcs://" + source[len("gs://") :]
-    return pl.read_parquet(source, **kwargs)
+
+    try:
+        # Polars added a crash bug in polars 0.20 that isn't fixed until polars 1.8.0; this fallback case handles
+        # that issue with a less efficient alternative that avoids their bug reading signed uris.
+        return pl.read_parquet(source, **kwargs)
+    except:
+        if isinstance(source, str):
+            if source.startswith("https://"):
+                buffer = BytesIO(requests.get(source).content)
+            else:
+                buffer = open(source, "rb")
+            try:
+                return pl.read_parquet(buffer, **kwargs)
+            except:
+                buffer.seek(0)
+                import pyarrow.parquet
+
+                pa_table = pyarrow.parquet.read_table(buffer, **kwargs)
+                pl_table = PyArrowToPolarsConverter(schema=pa_table.schema, use_fixed_size_list=True).convert(pa_table)
+                return pl_table
+        else:
+            raise
 
 
 if TYPE_CHECKING:
@@ -514,3 +648,210 @@ def arrow_ipc_to_record_batch(b: bytes) -> pa.RecordBatch:
 
 def pa_table_to_recordbatch(t: pa.Table) -> pa.RecordBatch:
     return pa.record_batch([c.combine_chunks() for c in t.columns], t.column_names)
+
+
+def estimate_row_size_bytes(
+    dtype: pa.DataType,
+    *,
+    average_string_size: int = 30,
+    average_bytes_size: int = 100,
+    average_list_length: int = 5,
+    null_fraction: float = 0.0,
+    float16_size: int = 2,
+) -> float:
+    """Estimate the average size of a row, given a schema. This is only slightly better than hardcoding logic based on row count, since it takes into account the width of the table.
+    However, if the estimates for string, bytes, or list sizes are off, then this estimate could be way off. It assumes no nulls"""
+    # Adding 0.125 to most fields to account for the nil bit
+    if pa.types.is_null(dtype):
+        return 0.125  # just the nil bit
+    if pa.types.is_boolean(dtype):
+        return 0.250  # bool bit + nil bit
+    if pa.types.is_int8(dtype) or pa.types.is_uint8(dtype):
+        return 1.125
+    if pa.types.is_int16(dtype) or pa.types.is_uint16(dtype):
+        return 2.125
+    if (
+        pa.types.is_int32(dtype)
+        or pa.types.is_uint32(dtype)
+        or pa.types.is_date32(dtype)
+        or pa.types.is_time32(dtype)
+        or pa.types.is_float32(dtype)
+    ):
+        return 4.125
+    if (
+        pa.types.is_int64(dtype)
+        or pa.types.is_uint64(dtype)
+        or pa.types.is_date64(dtype)
+        or pa.types.is_time64(dtype)
+        or pa.types.is_float64(dtype)
+        or pa.types.is_timestamp(dtype)
+        or pa.types.is_duration(dtype)
+    ):
+        return 8.125
+    if pa.types.is_float16(dtype):
+        return float16_size + 0.125
+    if pa.types.is_string(dtype) or pa.types.is_large_string(dtype):
+        return average_string_size + 0.125
+    if pa.types.is_binary(dtype) or pa.types.is_large_binary(dtype):
+        return average_bytes_size + 0.125
+    if pa.types.is_fixed_size_binary(dtype):
+        assert isinstance(dtype, pa.FixedSizeBinaryType)
+        return dtype.byte_width + 0.125
+    if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
+        assert isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType))
+        return (dtype.list_size if isinstance(dtype, pa.FixedSizeListType) else average_list_length) * (
+            estimate_row_size_bytes(
+                dtype.value_type,
+                average_string_size=average_string_size,
+                average_bytes_size=average_bytes_size,
+                average_list_length=average_list_length,
+                null_fraction=null_fraction,
+                float16_size=float16_size,
+            )
+        ) + 0.125
+    if pa.types.is_map(dtype):
+        assert isinstance(dtype, pa.MapType)
+        key_type = dtype.key_type
+        value_type = dtype.item_type
+        return (
+            estimate_row_size_bytes(
+                key_type,
+                average_string_size=average_string_size,
+                average_bytes_size=average_bytes_size,
+                average_list_length=average_list_length,
+                null_fraction=null_fraction,
+                float16_size=float16_size,
+            )
+            + estimate_row_size_bytes(
+                value_type,
+                average_string_size=average_string_size,
+                average_bytes_size=average_bytes_size,
+                average_list_length=average_list_length,
+                null_fraction=null_fraction,
+                float16_size=float16_size,
+            )
+        ) * average_list_length + 0.125
+    if pa.types.is_struct(dtype):
+        assert isinstance(dtype, pa.StructType)
+        size = 0.125
+        for i in range(dtype.num_fields):
+            f = dtype.field(i)
+            size += estimate_row_size_bytes(
+                f.type,
+                average_string_size=average_string_size,
+                average_bytes_size=average_bytes_size,
+                average_list_length=average_list_length,
+                null_fraction=null_fraction,
+                float16_size=float16_size,
+            )
+        return size
+
+    # If there's a type we didn't handle, assume 4 bytes
+    return 4.125
+
+
+def write_table_to_parquet(table: pa.Table, compression: str = "snappy") -> io.BytesIO:
+    buffer = io.BytesIO()
+    # From https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-parquet#input_file_requirements,
+    # row sizes need to be less than 50MB. I presume they meant to say row GROUP sizes in the docs, not individual row
+    # sizes. Assuming each row is approximately the same size, we can determine an appropriate row group size
+    if len(table) == 0:
+        # Doesn't really matter since there is no data!
+        target_row_group_size = 1
+    else:
+        avg_bytes_per_row = table.nbytes / len(table)
+        max_bytes_per_row_group = 50 * 1024 * 1024
+        avg_row_group_size = max_bytes_per_row_group / avg_bytes_per_row
+        # FIXME: Pick the target row group size via statistics. But I think targeting for 25MB row groups (i.e. half of the allowance)
+        # should be pretty safe
+        target_row_group_size = int(avg_row_group_size // 2)
+        # The default row group size is 1024*1024 (https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_table.html#pyarrow.parquet.write_table)
+        # Let's never exceed that
+        target_row_group_size = min(1024 * 1024, target_row_group_size)
+        # But let's always have at least one row
+        target_row_group_size = max(target_row_group_size, 1)
+    pq.write_table(
+        table,
+        buffer,
+        compression=compression,
+        use_compliant_nested_type=True,
+        row_group_size=target_row_group_size,
+    )
+    buffer.seek(0)
+    return buffer
+
+
+def describe_type_of_dataframe(frame: pl.DataFrame | pa.Table) -> str:
+    description = f"DataFrame shape=({frame.shape[0]} rows, {frame.shape[1]} cols)"
+
+    if isinstance(frame, pa.Table):
+        frame_columns = frame.column_names  # pyright: ignore[reportAttributeAccessIssue]
+    else:
+        frame_columns = frame.columns
+
+    if len(frame_columns) > 0:
+        if len(frame_columns) < 20:
+            description += f" with columns {repr(frame_columns)}"
+        else:
+            description += f" with selected columns {repr(frame_columns[:20])}"
+
+    return description
+
+
+def to_pydict(arrow_obj: pa.RecordBatch | pa.Table, /):
+    """Convert an arrow object to a python list using our optimized impl"""
+    return dict(zip(arrow_obj.column_names, (to_pylist(x) for x in arrow_obj.columns)))
+
+
+def to_pylist(arrow_obj: pa.Array | pa.ChunkedArray, /) -> list[object]:
+    """Convert an arrow object to a python list using our optimized impl"""
+    try:
+        from libchalk.utils import (  # pyright: ignore[reportMissingModuleSource,reportMissingImports]
+            arrow_array_to_pylist,
+        )
+    except ImportError:
+        return arrow_obj.to_pylist()
+    else:
+        if isinstance(arrow_obj, pa.ChunkedArray):
+            ans: list[object] = []
+            for chunk in arrow_obj.chunks:
+                if isinstance(chunk, pa.ExtensionArray):
+                    assert isinstance(
+                        arrow_obj.type, pa.ExtensionType
+                    ), "ExtensionArray must have an ExtensionType type."
+                    chunk = chunk.storage.cast(arrow_obj.type.storage_type)
+                ans.extend(arrow_array_to_pylist(chunk))
+            return ans
+        elif isinstance(arrow_obj, pa.ExtensionArray):
+            assert isinstance(arrow_obj.type, pa.ExtensionType), "ExtensionArray must have an ExtensionType type."
+            arrow_obj = arrow_obj.storage.cast(arrow_obj.type.storage_type)
+        return arrow_array_to_pylist(arrow_obj)
+
+
+def chunk_table(table: pa.Table, n: int) -> list:
+    """
+    Split a PyArrow Table into exactly `n` chunks.
+
+    Parameters:
+        table (pa.Table): The PyArrow Table to chunk.
+        n (int): The number of chunks to create.
+
+    Returns:
+        list: A list of PyArrow Tables, each forming a chunk.
+    """
+    if n <= 0:
+        raise ValueError("Number of chunks `n` must be greater than 0.")
+
+    if n == 1:
+        return [table]  # Return the whole table as a single chunk
+
+    total_rows = table.num_rows
+
+    if n >= total_rows:
+        # If n is greater than or equal to total rows, each chunk will have one row
+        return [table.slice(i, 1) for i in range(total_rows)]
+
+    chunk_size = math.ceil(total_rows / n)
+    chunks = [table.slice(i, chunk_size) for i in range(0, total_rows, chunk_size)]
+
+    return chunks
